from threemystic_cloud_client.cloud_providers.aws.config.base_class.base import cloud_client_aws_config_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers
import configparser

class cloud_client_aws_config_step_2(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_aws_config_step_2", *args, **kwargs)

  def step_done(self, *args, **kwargs):
    
    self.update_provider_config_completed(status= True)
    self._setup_another_config()


  def step(self, is_new_config, *args, **kwargs):

    if not super().step():
      return

    if is_new_config:
      self.__step_new()
      return
    
    self.__step_existing()

  def __step_new(self, *args, **kwargs):

    if not super().step():
      return
    
    profile_data = {}
    response = self.get_common().generate_data().generate(
      generate_data_config = self.get_common().helper_type().dictionary().merge_dictionary([
        {}, 
        self.__get_profile_data_genrationn(),
        {
          "auth_method": {
            "validation": lambda item: self.get_common().helper_type().string().set_case(string_value= item, case= "lower") in self.valid_auth_options(),
            "messages":{
              "validation": f"Valid Options: {self.valid_auth_options()}",
            },
            "conversion": lambda item: self.get_common().helper_type().string().trim(string_value= self.get_common().helper_type().string().trim(string_value= self.get_common().helper_type().string().set_case(string_value= item, case= "lower"))) if item is not None else None,
            "desc": f"Which provider authentication\nvalid options are {self.valid_auth_options()}",
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": False
          }
        }
      ])
    )

    if response is None:
      return
    
    profile_name = self.__get_response_item(key= "profile_name", response= response)
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_name):
      print("Profile Name was empty.")
      return     
      
    profile_data["auth_method"] = self.__get_response_item(key= "auth_method", response= response)   
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_data["auth_method"]):
      print("Auth Method was empty.")
      return 

    if(self.config_profile_name_exists(profile_name= profile_name)):
      response_existing = self.get_common().generate_data().generate(
        generate_data_config = {
          "update_existing": {
            "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
            "allow_empty": True,
            "messages":{
              "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}\nValid options for No are: {self.get_common().helper_type().bool().is_false_values()}",
            },
            "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
            "desc": f"Profile Name {profile_name} already exists.\nDo you want to update the existing profile?\nValid optiond for yes: {self.get_common().helper_type().bool().is_true_values()}{self.__get_existing_text(exiting_value= profile_data.get('default_profile'))}",
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "default": False,
            "optional": True
          }
        }
      )
      if(self.__get_response_item(key= "update_existing", response= response_existing) == False):
        print("Please rerun config and use a different profile name")
        return

      profile_data = self.get_config_profile_name(profile_name= profile_name)
      profile_data["auth_method"] = self.__get_response_item(key= "auth_method", response= response) 
    
    self.__step_process_sso(
      profile_name= profile_name,
      profile_data= profile_data
    )
  
  def __step_existing(self, *args, **kwargs):
    if not self.has_config_profiles():
      print("No existing profiles please run new")
      return
    
    response = self.get_common().generate_data().generate(
      generate_data_config = self.__get_profile_data_genrationn()
    )
    
    profile_name = self.__get_response_item(key= "profile_name", response= response)

    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_name):      
      print(f"Profile Name was not valid: {profile_name}")
      return
    
    profile_data = self.get_config_profile_name(profile_name= profile_name)
    
    if profile_data is None:
      print("Profile not found")
      return

    if profile_data.get("auth_method") == "sso":
      self.__step_process_sso(
        profile_name= profile_name,
        profile_data= profile_data
      )

  def __get_profile_data_genrationn(self, *args, **kwargs):
    return {
      "profile_name": {
        "validation": lambda item: self.get_common().helper_type().regex().get(pattern= "^[a-z][a-z0-9_-]{1,}$").fullmatch(self.get_common().helper_type().string().trim(string_value= str(item))) if item is not None else False,
        "messages":{
          "validation": f"It should be alphanumeric and can have underscores or dashes and must start with a letter and be at lest 2 characters ^[a-z][a-z0-9_-]\{1,}$",
        },
        "conversion": lambda item: self.get_common().helper_type().string().trim(string_value= self.get_common().helper_type().string().set_case(string_value= item, case= "lower")) if item is not None else None,
        "desc": f"Profile Name (it will be converted to all lowercase)\nRequirments: alphanumeric and can have underscores or dashes and must start with a letter and be at lest 2 characters ^[a-z][a-z0-9_-]\{1,}$",
        "handler": generate_data_handlers.get_handler(handler= "base"),
        "optional": False
      }
    }
  
  def __get_response_item(self, key, response, *args, **kwargs):
    if response is None:
      return None

    if response.get(key) is not None:
      return self.__get_response_item(key= key, response= response.get(key))
    
    return self.get_common().helper_type().string().set_case(string_value= response.get("formated"), case= "lower")   
    
  

 
  
  def __step_process_sso_valid_profile(self, profile_name, existing_profile, *args, **kwargs):

    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_name):
      return False if self.get_common().helper_type().string().is_null_or_whitespace(string_value= existing_profile) else True
      
    
    profile_name = self.get_common().helper_type().string().trim(string_value= profile_name)  
    config_parser = configparser.ConfigParser()
    if(self.get_common().helper_path().path_exists(path= self.get_aws_user_path_config())):
      with self.get_common().helper_path().expandpath_user(path= self.get_aws_user_path_config()).open(mode="r") as config_file:
        config_parser.read_file(config_file)
    else:
      print()
      print()
      print("*********************************************************")
      print(f"AWS config path not found: {self.get_aws_user_path_config()}")
      print("setup will not work until aws is configured")
      print("*********************************************************")
      print()
      print()
      
      return True
    return config_parser.has_section(f"profile {profile_name}")

  def get_default_profile_setting(self, profile_name, profile_data):
    if profile_data is not None:
      return profile_data.get("default_profile") == True
    
    if self.get_default_profile() is None:
      return True

    existing_profile = self.get_config_profile_name(profile_name= profile_name)
    if(existing_profile is not None):
      return existing_profile.get("default_profile") == True
    
    
    return False
    

  def __get_existing_text(self, exiting_value):
    return (f"\n(If empty it will use the existing: {exiting_value})"
      if not self.get_common().helper_type().string().is_null_or_whitespace(string_value=exiting_value) else ""
    )

  def __step_process_sso(self, profile_name = None, profile_data = {}, *args, **kwargs):
    if(profile_data is None):
      profile_data = {}

    existing_sso_profile_name = profile_data.get("sso_profile_name")
    
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "use_cli_profile": {
          "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
          "allow_empty": True,
          "messages":{
            "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}\nValid options for No are: {self.get_common().helper_type().bool().is_false_values()}",
          },
          "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
          "desc": f"Use preconfigured aws cli sso profile (created with aws configure sso --profile <profile_name>)\nValid optiond for yes: {self.get_common().helper_type().bool().is_true_values()}{self.__get_existing_text(exiting_value= profile_data.get('use_cli_profile'))}",
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "default": profile_data.get("use_cli_profile") == True,
          "optional": False
        },
        "sso_profile_name": {
          "validation": lambda item: self.__step_process_sso_valid_profile(profile_name= item, existing_profile= existing_sso_profile_name),
          "allow_empty": False,
          "skip": lambda item: item.get("use_cli_profile").get("formatted") if item is not None and item.get("use_cli_profile") is not None else False,
          "messages":{
            "validation": f"Could not find profile.",
          },
          "conversion": lambda item: self.get_common().helper_type().string().trim(string_value= item),
          "desc": f"Enter the SSO Profile Name{self.__get_existing_text(exiting_value= existing_sso_profile_name)}",
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "default": profile_data.get("sso_profile_name"),
          "optional": not self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_data.get("sso_profile_name"))
        },
        "sso_start_url": {
          "validation": lambda item: item,
          "skip": lambda item: not item.get("use_cli_profile").get("formatted") if item is not None and item.get("use_cli_profile") is not None else True,
          "messages":{},
          "conversion": lambda item: self.get_common().helper_type().string().trim(item),
          "desc": f"Enter the SSO start url (ex: https://<aws_id>.awsapps.com/start",
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "default": profile_data.get("sso_start_url"),
          "optional": not self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_data.get("sso_start_url"))
        },
        "sso_region": {
          "validation": lambda item: item,
          "skip": lambda item: not item.get("use_cli_profile").get("formatted") is not True if item is not None and item.get("use_cli_profile") is not None else True,
          "allow_empty": True,
          "messages":{},
          "conversion": lambda item: self.get_common().helper_type().string().trim(item),
          "desc": f"Enter the default region to use{self.__get_existing_text(exiting_value= profile_data.get('sso_region'))}",
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "default": profile_data.get("sso_region") if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_data.get("sso_region")) else "us-east-1",
          "optional": False
        },
        "sso_account_id": {
          "validation": lambda item: self.get_common().helper_type().regex().get(pattern= "^[0-9]{12,}$").fullmatch(str(item)) if item is not None else False,
          "skip": lambda item: not item.get("use_cli_profile").get("formatted") if item is not None and item.get("use_cli_profile") is not None else True,
          "allow_empty": True,
          "messages":{
            "validation": f"It should be a 12 digit string (if its under 12 characters it should have leading zeros) ex. 000000000001",
          },
          "conversion": lambda item: item,
          "desc": f"Enter the organization account id (main account id)\n It should be 12 numeric characters.{self.__get_existing_text(exiting_value= profile_data.get('sso_account_id'))}",
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "default": profile_data.get("sso_account_id"),
          "optional": False
        },
        "sso_role_name": {
          "validation": lambda item: item,
          "skip": lambda item: not item.get("use_cli_profile").get("formatted") if item is not None and item.get("use_cli_profile") is not None else True,
          "allow_empty": True,
          "messages":{},
          "conversion": lambda item: self.get_common().helper_type().string().trim(item),
          "desc": f"Enter the role to use{self.__get_existing_text(exiting_value= profile_data.get('sso_account_id'))}",
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "default": profile_data.get("sso_role_name"),
          "optional": False
        },
        "output": {
          "validation": lambda item: item,
          "skip": lambda item: not item.get("use_cli_profile").get("formatted") if item is not None and item.get("use_cli_profile") is not None else True,
          "allow_empty": True,
          "messages":{},
          "conversion": lambda item: self.get_common().helper_type().string().trim(item),
          "desc": f"Enter a valid output format. For a full list goto:\nhttps://docs.aws.amazon.com/cli/latest/userguide/cli-usage-output-format.html{self.__get_existing_text(exiting_value= profile_data.get('output'))}",
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "default": profile_data.get("output") if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_data.get("output")) else "json",
          "optional": False
        },        
        "default_profile": {
          "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
          "allow_empty": True,
          "messages":{
            "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}\nValid options for No are: {self.get_common().helper_type().bool().is_false_values()}",
          },
          "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
          "desc": f"Is this the default profile 3mystic apps should use when profile is not passed. You can only have one profile,\nValid optiond for yes: {self.get_common().helper_type().bool().is_true_values()}{self.__get_existing_text(exiting_value= profile_data.get('default_profile'))}",
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "default": self.get_default_profile_setting(profile_name= profile_name, profile_data= profile_data),
          "optional": False
        }
      }
    )
    
    if response is None:
      return

    for key, item in response.items():
      profile_data[key] = item.get("formated") if item is not None else ""
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value=profile_data.get("sso_profile_name")):
      profile_data["sso_profile_name"] = existing_sso_profile_name
    
    self.update_config_profile(profile_name= profile_name, profile_data= profile_data)
    print(f"Profile ({profile_name} saved/updated)")
    self.step_done()
    

    
  



from threemystic_cloud_client.cloud_providers.azure.action_test.base_class.base import cloud_client_azure_test_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers


class cloud_client_azure_test_step_1(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_azure_test", *args, **kwargs)
    

  def step(self, *args, **kwargs):
    if not super().step( *args, **kwargs):
      return
    
    from threemystic_cloud_client.cloud_client import cloud_client
    azure_client = cloud_client(logger= self.get_common().get_logger(), common=self.get_common()).client(
      provider= "azure"
    )
    
    print(f"You have the following tenants:")
    for tenant in azure_client.get_tenants():
      print(f"{azure_client.get_tenant_id(tenant= tenant)}")
    
    print(f"You have the following accounts:")
    for account in azure_client.get_accounts():
      print(f"{azure_client.get_account_id(account= account)}:{azure_client.get_account_name(account= account)}")

    print()
    print()
    print("--------------------------------")
    print()
    print()
    print("If not all your accounts are listing you can try the commands below. ")
    print("If accounts are still missing please log into the portal and verify all the tenants. It is possible they are missing from the CLI.")
    for tenant in azure_client.get_tenants():
      print(f"az login --tenant {azure_client.get_tenant_id(tenant= tenant)}")

    print(f"az account list --refresh")
    
  



from threemystic_cloud_client.cloud_providers.aws.client.base_class.base import cloud_client_aws_client_base as base
import sys
import os
import boto3
from polling2 import TimeoutException, poll as poll2
    
class cloud_client_aws_client_sso(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_aws_client_sso", *args, **kwargs)

    if(self.get_common().helper_type().string().set_case(string_value= self.get_profile()["profile_data"]["auth_method"], case= "lower") != "sso"):
      raise self.get_common().exception().exception(
          exception_type = "generic"
        ).type_error(
          logger = self.get_common().get_logger(),
          name = "Profile Authentication Invalid",
          message = f"Profile Authentication Invalid - Cloud Client Profile: {self.get_profile()['profile_name']}"
        )
    
    if(self.get_profile()['profile_data']['use_cli_profile'] and not self._has_aws_sso_config_profile()):
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = f"AWSCLI SSO Profile NOT FOUND",
        message = f"AWSCLI SSO Profile NOT FOUND - Cloud Client Profile: {self.get_profile()['profile_name']}\nPlease run the following command to configure aws cli:\naws configure sso --profile {self.get_profile()['profile_data']['sso_profile_name']}"
      )
  
  def __get_aws_cli_config(self, refresh= False, *args, **kwargs):
    if(hasattr(self, "_aws_config") and (not refresh)):
      return self._aws_config
    
    self._aws_config = self.get_common().helper_config().load(
      config_type = "config",
      path= self.get_aws_user_path_config()
    )
    return self.__get_aws_cli_config(*args, **kwargs)
  
  def __unset_aws_cli_config(self, *args, **kwargs):
    if(hasattr(self, "_aws_config")):
      delattr(self, "_aws_config")
  
  
  def _has_aws_sso_config_profile(self, refresh = False, *args, **kwargs):

    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= self.get_profile()['profile_data']['sso_profile_name']) and self.__get_aws_cli_config(refresh= refresh).has_section(f"profile {self.get_profile()['profile_data']['sso_profile_name']}"):
      return True
    
    return False
  
  def _get_aws_sso_config_profile(self, refresh = False, *args, **kwargs):
    
    if(hasattr(self, "_aws_config_profile") and (not refresh)):
      return self._aws_config_profile
    
    if(not self._has_aws_sso_config_profile(refresh= refresh)):
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = f"NOT FOUND",
        message = f"AWS SSO PRofile NOT FOUND"
      )
    
    self._aws_config_profile = self.__get_aws_cli_config()[f"profile {self.get_profile()['profile_data']['sso_profile_name']}"]
    self.__unset_aws_cli_config()
    return self._get_aws_sso_config_profile(*args, **kwargs)
  
  def _get_sso_profile_credentials(self, refresh = False, *args, **kwargs):   
    if(hasattr(self, "_aws_config_profile_credentials") and (not refresh)):
      if len(self._aws_config_profile_credentials) > 0:
        return self._aws_config_profile_credentials
    
    cache_key = f'{self.get_common().encryption().hash(hash_method="sha1").generate_hash(data= self._get_aws_sso_config_profile(refresh= refresh)["sso_start_url"])}.json' 
    
    sso_token_file = self.get_aws_user_path().joinpath('sso', 'cache', cache_key)
    
    self._aws_config_profile_credentials = {}
    if not self.get_common().helper_path().is_file(sso_token_file):
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = f"NOT FOUND",
        message = f"Token File not found or not valid file: {sso_token_file}"
      )
    
    with sso_token_file.open(mode="r") as sso_cache:
      self._aws_config_profile_credentials = self.get_common().helper_json().loads(data= sso_cache.read())
    
    return self._get_sso_profile_credentials()
  
  def __internal_load_base_configs_ssoprofile(self, *args, **kwargs):
    self._get_sso_profile_credentials()

  def _get_session_accesstoken(self, refresh = False, *args, **kwargs):
    if self.session_expired():
      self.ensure_session()

    if(hasattr(self, "_aws_config_profile_credentials_accesstoken") and (not refresh)):
      if(not self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._aws_config_profile_credentials_accesstoken) ):
        return self._aws_config_profile_credentials_accesstoken
    
    self._aws_config_profile_credentials_accesstoken = self._get_sso_profile_credentials().get("accessToken")
    return self._get_session_accesstoken(refresh= refresh)
  
  def _get_session_expires(self, refresh = False, *args, **kwargs):
    if(hasattr(self, "_aws_config_profile_credentials_expires") and (not refresh)):
      return self._aws_config_profile_credentials_expires

    if(self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._get_sso_profile_credentials(refresh= refresh).get("expiresAt")) ):
      return (self.get_common().helper_type().datetime().get() - self.get_common().helper_type().datetime().time_delta(seconds= 300))
    
    self._aws_config_profile_credentials_expires = (self.get_common().helper_type().datetime().convert_utc(
      dt= self.get_common().helper_type().datetime().parse_iso(iso_datetime_str= self._get_sso_profile_credentials().get("expiresAt")))
      - self.get_common().helper_type().datetime().time_delta(seconds= 300)
    )

    return self._get_session_expires()
    
  def _session_expired(self, refresh = False, *args, **kwargs):

    return self.get_common().helper_type().datetime().is_token_expired_now(compare_datetime= self._get_session_expires(refresh= refresh))
    
    
  def authenticate_session(self, force_quiet = False, *args, **kwargs):
    if(self.get_profile()['profile_data']['use_cli_profile']):
      self.__aws_sso_login_profile(force_quiet= force_quiet)
      return
    
    raise self.get_common().exception().exception(
      exception_type = "generic"
    ).type_error(
      logger = self.get_common().get_logger(),
      name = f"NOT SUPPORTED",
      message = f"Cloud Client Profile not loaded"
    )

  def __aws_sso_login_profile(self, force_quiet = False, *args, **kwargs):
    if(not self._has_aws_sso_config_profile()):
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = f"NOT FOUND",
        message = f"Cloud Client Profile not found - Login"
      )
    
    ssologin_call = f"aws sso login --profile {self.get_profile()['profile_data']['sso_profile_name']}"
    if force_quiet:
      ssologin_call = f'{ssologin_call} 1>/dev/null'
    if os.system(ssologin_call) != 0:
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = f"NOT AUTHENTICATED",
        message = f"Cloud Client Profile could not authenticate SSO Profile"
      )

    try:
      logged_in = poll2(
        lambda: not self.session_expired(refresh = True),
        ignore_exceptions=(Exception,),
        timeout=self.get_aws_poll_login(),
        step=0.1
      )
      if logged_in is not None:      
        return logged_in
      
      raise self.get_common().exception().exception(
          exception_type = "generic"
        ).type_error(
          logger = self.get_common().get_logger(),
          name = f"NOT SUPPORTED",
          message = f"Could not authenticate SSO.\nPlease check cli version aws --version and make sure you are using v2 then please run aws sso login --profile {self.get_profile()['profile_data']['sso_profile_name']}"
        )
    except Exception as err:
      raise self.get_common().exception().exception(
          exception_type = "generic"
        ).type_error(
          logger = self.get_common().get_logger(),
          name = f"NOT SUPPORTED",
          message = f"Could not authenticate SSO.\nPlease check cli version aws --version and make sure you are using v2 then please run aws sso login --profile {self.get_profile()['profile_data']['sso_profile_name']}"
        )


  def _load_base_configs(self, *args, **kwargs):
    if(self.get_profile()['profile_data']['use_cli_profile']):
      self.__internal_load_base_configs_ssoprofile()
      return
    
    raise self.get_common().exception().exception(
      exception_type = "generic"
    ).type_error(
      logger = self.get_common().get_logger(),
      name = f"NOT SUPPORTED",
      message = f"Cloud Client Profile not loaded"
    )

  def get_default_rolename(self, *args, **kwargs):
    if(self.get_profile()['profile_data']['use_cli_profile']):
      return self._get_aws_sso_config_profile()["sso_role_name"]
    
    raise self.get_common().exception().exception(
      exception_type = "generic"
    ).type_error(
      logger = self.get_common().get_logger(),
      name = f"NOT SUPPORTED",
      message = f"Unknown default role"
    )

  def get_default_region(self, *args, **kwargs):
    if(self.get_profile()['profile_data']['use_cli_profile']):
      if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._get_aws_sso_config_profile()["region"]):
        return self._get_aws_sso_config_profile()["region"]
      
      if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._get_aws_sso_config_profile()["sso_region"]):
        return self._get_aws_sso_config_profile()["sso_region"]
    
    raise self.get_common().exception().exception(
      exception_type = "generic"
    ).type_error(
      logger = self.get_common().get_logger(),
      name = f"NOT SUPPORTED",
      message = f"Unknown default region"
    )
            
  def get_default_account(self):
    return self.make_account(Id = self._get_aws_sso_config_profile()["sso_account_id"])
  
  def __get_sso_boto_session(self):
    return boto3.Session(
      profile_name = self.get_profile()['profile_data']['sso_profile_name']
    )
  
  def _assume_role(self, account = None, region = None, role = None, refresh= False, *args, **kwargs):
        
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= role):
      role = self.get_default_rolename()
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= region):
      region = self.get_default_region()

    if account is None or (self.get_common().helper_type().general().is_type(obj= account, type_check= str) and self.get_common().helper_type().string().is_null_or_whitespace(string_value= account)):
      account = self.get_default_account()

    if not refresh and self._get_assumed_role_credentials().get(self._get_assumed_role_credentials_key(account= account, role= role)) is not None:
      expiration = self._auto_parse_aws_expiration(
        expiration= self._get_assumed_role_credentials()[self._get_assumed_role_credentials_key(account= account, role= role)]["roleCredentials"]["expiration"]
      )

      if not self.get_common().helper_type().datetime().is_token_expired_now(compare_datetime= expiration):
        return self._get_assumed_role_credentials()[self._get_assumed_role_credentials_key(account= account, role= role)]["roleCredentials"]
    
    self._get_assumed_role_credentials()[self._get_assumed_role_credentials_key(account= account, role= role)] = self.__get_sso_boto_session().client('sso').get_role_credentials(
      roleName=role,
      accountId=self.get_account_id(account= account),
      accessToken=self._get_session_accesstoken()
    )

    return self._assume_role(
      account = account, region = region, role = role, *args, **kwargs
    )
                
  
  def get_main_account_id(self, *args, **kwargs):
    return self.get_account_id(account= self.get_default_account())

  def get_organization_account_id(self, *args, **kwargs):
    return self.get_account_id(account= self.get_default_account())



from threemystic_cloud_client.cli.actions.base_class.base import cloud_client_action_base as base


class cloud_client_test(base):
  def __init__(self, *args, **kwargs):
    super().__init__(action= "--test,-t", *args, **kwargs)


  def _process_provider_aws(self, *args, **kwargs):
    from threemystic_cloud_client.cloud_providers.aws  import cloud_client_aws as client
    client(common= self._cloud_client.get_common()).action_test()


  def _process_provider_azure(self, *args, **kwargs):
    from threemystic_cloud_client.cloud_providers.azure import cloud_client_azure as client
    client(common= self._cloud_client.get_common()).action_test()      

      

  



from threemystic_cloud_client.cloud_providers.base_class.base import cloud_client_provider_base as base
import os
from botocore import session as botocore_session, credentials as botocore_credentials
from botocore.config import Config as botocore_config_config
from boto3 import Session as boto_session
from botocore.exceptions import ClientError, SSLError
from polling2 import TimeoutException, poll as poll2
import time
from random import randint

class cloud_client_provider_aws_base(base):
  def __init__(self, *args, **kwargs):
    # https://github.com/boto/botocore/issues/2705 
    # This update should be temporary until boto version 1.28 is released
    # os.environ["BOTO_DISABLE_COMMONNAME"] = "true"

    super().__init__(provider= "aws", *args, **kwargs)   
    self.links = {
      "cli_doc_link": "https://docs.aws.amazon.com/cli/latest/userguide/getting-started-install.html",
      "ssm_doc_link": "https://docs.aws.amazon.com/systems-manager/latest/userguide/session-manager-working-with-install-plugin.html",
      "saml2aws_doc_link": "https://github.com/Versent/saml2aws"
    }

  def get_resource_name(self, resource, *args, **kwargs):
    if resource is None:
      return None

    if self.get_common().helper_type().general().is_type(obj= resource, type_check= str):
      return resource
    
    if self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      for key in resource.keys():
        if self.get_common().helper_type().string().set_case(string_value= key, case= "lower") == "name":
          return self.get_resource_name(resource= resource.get(key))
        
      resource_tags = self.get_resource_tags_as_dictionary(resource= resource)
      if resource_tags is None:
        return None
      if len(resource_tags) < 1:
        return None
      
      return self.get_resource_name(resource= resource)

    return None
  
  def get_account_name(self, account):
    if account is None:
      return None
    if self.get_common().helper_type().general().is_type(obj= account, type_check= str):
      return account.strip()

    if account.get("Name"):
      return account["Name"].strip()
    
    if account.get("accountName"):
      self.get_common().get_logger().warning("accountName will be depreciated use Name")
      return account["accountName"].strip()
    
    raise self.get_common().exception().exception(
      exception_type = "generic"
    ).not_implemented(
      logger = self.get_common().get_logger(),
      name = "account",
      message = f"Unknown account object: {account}."
    )

  def get_account_id(self, account):
    if account is None:
      return None
    if self.get_common().helper_type().general().is_type(obj= account, type_check= str):
      return self.get_common().helper_type().string().trim(string_value=self.get_common().helper_type().string().set_case(string_value= account, case= "lower"))
    
    if account.get("Id"):
      return self.get_common().helper_type().string().trim(string_value=self.get_common().helper_type().string().set_case(string_value= account["Id"], case= "lower"))

    if account.get("accountId"):
      self.get_common().get_logger().warning("accountId will be depreciated use Id")
      return self.get_common().helper_type().string().trim(string_value=self.get_common().helper_type().string().set_case(string_value= account["accountId"], case= "lower"))
    
    raise self.get_common().exception().exception(
      exception_type = "generic"
    ).not_implemented(
      logger = self.get_common().get_logger(),
      name = "account",
      message = f"Unknown account object: {account}."
    )
  
  def make_account(self, **kwargs):    
    account = {}
    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= kwargs.get("accountId")):
      self.get_common().get_logger().warning("accountId will be depreciated use Id")
      account["accountId"] = self.get_common().helper_type().string().trim(string_value=self.get_common().helper_type().string().set_case(string_value= kwargs.get("accountId"), case= "lower"))
      account["Id"] = self.get_common().helper_type().string().trim(string_value=self.get_common().helper_type().string().set_case(string_value= kwargs.get("accountId"), case= "lower"))

    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= kwargs.get("Id")):
      account["Id"] = self.get_common().helper_type().string().trim(string_value=self.get_common().helper_type().string().set_case(string_value= kwargs.get("Id"), case= "lower"))

    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= kwargs.get("accountName")):
      self.get_common().get_logger().warning("accountName will be depreciated use Name")
      account["accountName"] = self.get_common().helper_type().string().trim(string_value=kwargs.get("accountName"))
      account["Name"] = self.get_common().helper_type().string().trim(string_value=kwargs.get("accountName"))

    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= kwargs.get("Name")):
      account["Name"] = self.get_common().helper_type().string().trim(string_value=kwargs.get("Name"))
    
    return account
  
  @property
  def default_region_resources(self):
    if hasattr(self, "_default_region_resources"):
      return self._default_region_resources
    
    return ["us-east-1"]
  
  @default_region_resources.setter
  def _set_default_region_resources(self, value):
    self._default_region_resources = value

  def get_resource_general_arn(self, resource_type = None, resource_type_sub = None, account_id = None, region = None, resource_id = None, data_item = None, **kwargs ):
    if data_item is not None:
      lower_keys = [key.lower() for key in data_item.keys()]
      if "arn" in lower_keys:
        arn_key = list(data_item.keys())[lower_keys.index("arn")]
        return self.get_common().helper_type().string().set_case(string_value= data_item[arn_key], case= "lower")

    return self.get_common().helper_type().string().set_case(string_value= f'arn:aws:{resource_type}:{region}:{account_id}:{resource_type_sub}/{resource_id}', case= "lower")
  
  def _get_boto_client_key(self, client, account = None, region = None, *args, **kwargs):
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= region):
      region = self.get_default_region()

    if account is None or (self.get_common().helper_type().general().is_type(obj= account, type_check= str) and self.get_common().helper_type().string().is_null_or_whitespace(string_value= account)):
      account = self.get_default_account()

    return f"{client}-{(self.get_account_id(account))}-{region}" 
  
  def _get_created_boto_clients(self, *args, **kwargs):
    if(hasattr(self, "_created_boto_clients")):
      return self._created_boto_clients
    
    self._created_boto_clients = {}
    return self._get_created_boto_clients()
  
  def _get_boto_session_key(self, account = None, role = None, region = None, profile = None, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= role):
      role = self.get_default_rolename()
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= region):
      region = self.get_default_region()

    if account is None or (self.get_common().helper_type().general().is_type(obj= account, type_check= str) and self.get_common().helper_type().string().is_null_or_whitespace(string_value= account)):
      account = self.get_default_account()

    cached_key = f"{self.get_account_id(account)}-{region}-{role}"
    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile):
      cached_key = f"{cached_key}-{profile}"

      return cached_key
  
  def _get_created_boto_sessions(self, *args, **kwargs):
    if(hasattr(self, "_created_boto_sessions")):
      return self._created_boto_sessions
    
    self._created_boto_sessions = {}
    return self._get_created_boto_sessions()
  
  def _get_assumed_role_credentials_key(self, account, role, *args, **kwargs):    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= role):
      role = self.get_default_rolename()

    if account is None or (self.get_common().helper_type().general().is_type(obj= account, type_check= str) and self.get_common().helper_type().string().is_null_or_whitespace(string_value= account)):
      account = self.get_default_account()

    return f'{self.get_account_id(account= account)}_{role}'

  def _get_assumed_role_credentials(self, unset = False, *args, **kwargs):
    if(hasattr(self, "_assumed_role_credentials")):
      if unset:
        current = self._assumed_role_credentials
        delattr(self, "_assumed_role_credentials")
        return current
      
      return self._assumed_role_credentials
    
    self._assumed_role_credentials = {}
    return self._get_assumed_role_credentials()
    
  def get_profile(self, *args, **kwargs):
    if(not hasattr(self, "_profile")):
      self._set_profile()
    
    if(self._profile is None):
      raise self.get_common().exception().exception(
          exception_type = "generic"
        ).type_error(
          logger = self.get_common().get_logger(),
          name = "Cloud Client Profile",
          message = f"Profile is None"
        )
    
    return self._profile
  
  def _set_profile(self, profile_data = None, *args, **kwargs):
    if profile_data is None:
      self._profile = self.get_default_profile()
      return
      
    self._profile = profile_data
  
  async def async_general_boto_call_array(self, *args, **kwargs):
    return self.general_boto_call_array(*args, **kwargs)
  
  async def async_general_boto_call_single(self, *args, **kwargs):
    return self.general_boto_call_single(*args, **kwargs)
  
  def general_boto_call_single(self, *args, **kwargs):
    data = self.general_boto_call_array(*args, **kwargs)
    if data is None or len(data) < 1:
      return None
    
    return data[0]
  
  def general_boto_call_array(self, boto_call, boto_params, boto_key, boto_nextkey, retryCount = 10, verbose = False, boto_nextkey_param = None, error_codes_return = None, error_codes_continue = None, error_codes_raise = None, logger = None):
    return_data = []

    if boto_key is not None and self.get_common().helper_type().general().is_type(obj=boto_key, type_check= str):
      local_boto_key = boto_key
      boto_key = lambda item: item[local_boto_key] if item.get(local_boto_key) is not None else []

    if boto_nextkey_param is None:
      boto_nextkey_param = boto_nextkey

    if error_codes_continue is not None:
      error_codes_continue = [self.get_common().helper_type().string().set_case(string_value= code, case= "lower") for code in error_codes_continue]

    if error_codes_return is not None:
      error_codes_return = [self.get_common().helper_type().string().set_case(string_value= code, case= "lower") for code in error_codes_return]

    if error_codes_raise is not None:
      error_codes_raise = [self.get_common().helper_type().string().set_case(string_value= code, case= "lower") for code in error_codes_raise]

    boto_response = None
    while True:
      currentAttempt = 0
      slowdown_count = 0
      
      while currentAttempt < retryCount:
        currentAttempt += 1
        try:
          if boto_params is not None:
            boto_response = boto_call(boto_params)
            break
          
          boto_response = boto_call()
          break
        except ClientError as err:
          if error_codes_raise is not None and self.get_common().helper_type().string().set_case(string_value= err.response["Error"]["Code"], case= "lower") in error_codes_raise:
            raise self.get_common().exception().exception(
              exception_type = "generic"
            ).type_error(
              logger = self.get_common().get_logger(),
              name = "General Boto Call Raise",
              message = f"error_codes_raise - {err.response['Error']['Code']} - {err.response['ResponseMetadata']['RequestId']} - {err.response['Error']['Message']}",
              exception= err,
              log_as_warning = True
            )

          if error_codes_continue is not None and self.get_common().helper_type().string().set_case(string_value= err.response["Error"]["Code"], case= "lower") in error_codes_continue:
            continue

          if error_codes_return is not None and self.get_common().helper_type().string().set_case(string_value= err.response["Error"]["Code"], case= "lower") in error_codes_return:
            return return_data
          
          if self.get_common().helper_type().string().set_case(string_value= err.response["Error"]["Code"], case= "lower") == "accessdeniedexception":
            raise self.get_common().exception().exception(
              exception_type = "generic"
            ).type_error(
              logger = self.get_common().get_logger(),
              name = "General Boto Call accessdeniedexception",
              message = f"accessdeniedexception - {err.response['Error']['Code']} - {err.response['ResponseMetadata']['RequestId']} - {err.response['Error']['Message']}",
              exception= err,
              log_as_warning = True
            )
          
          if err.response['Error']["Code"] == 'SlowDown':
            time.sleep(self.get_common().helper_type().requests().expodential_backoff_wait(attempt= currentAttempt, auto_sleep= False))
            if slowdown_count < 5:
              currentAttempt-=1
            continue

          if verbose:
            self.get_common().get_logger().exception(
              msg=f"Params:{boto_params} - err: {err}",
              extra= {
                "main_exception": err
              }
            )

          if currentAttempt >= retryCount:
            boto_response = None
            raise self.get_common().exception().exception(
              exception_type = "generic"
            ).type_error(
              logger = self.get_common().get_logger(),
              name = "General Boto Call err retry ",
              message = f"err_retrycount - {currentAttempt} - {retryCount}",
              exception= err,
              log_as_warning = True
            )
            
          if currentAttempt > 2:
            self.get_common().get_logger().exception(msg= "Error with call: {}".format(err), exc_info= err)
            if verbose:
              self.get_common().get_logger().info(msg= "Error with call: {}".format(err))

          time.sleep(self.get_common().helper_type().requests().expodential_backoff_wait(attempt= currentAttempt, auto_sleep= False))
          continue
        except SSLError as err:
          if currentAttempt >= retryCount:
            boto_response = None
            raise self.get_common().exception().exception(
              exception_type = "generic"
            ).type_error(
              logger = self.get_common().get_logger(),
              name = "General Boto Call err retry ",
              message = f"SSLError - err_retrycount - {currentAttempt} - {retryCount}",
              exception= err,
              log_as_warning = True
            )
          continue


      if boto_response is None:
        return [ ]
      
      if boto_key is None:
        return_data.append(boto_response)
        
      if boto_key is not None:
        if not self.get_common().helper_type().general().is_type(obj= boto_key(boto_response), type_check= list):
          return [ boto_key(boto_response) ]

        return_data += boto_key(boto_response)

      if (boto_nextkey is None) or (boto_nextkey is not None and self.get_common().helper_type().string().is_null_or_whitespace(string_value= boto_response.get(boto_nextkey))):
        return return_data

      if boto_params is not None:
        boto_params[boto_nextkey_param] = boto_response.get(boto_nextkey)
        continue

      boto_params= {
        boto_nextkey_param: boto_response.get(boto_nextkey)
      }

  def get_organization_account(self, *args, **kwargs):
    if(hasattr(self, "_organization_account")):
      if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._organization_account):
        return self._organization_account
      
    
    self._organization_account = self.make_account(Id = self.get_organization_account_id())
    return self.get_organization_account()

  def assume_role(self, *args, **kwargs):
    self.ensure_session()
    if kwargs.get("account") is not None and self.get_common().helper_type().general().is_type(obj= kwargs["account"], type_check= str):
      kwargs["account"] = self.make_account(Id= kwargs["account"]) 

    return self._assume_role(**kwargs)
  
  def session_expired(self, refresh = False, *args, **kwargs):
    return self._session_expired(refresh= refresh, *args, **kwargs)
  
  def ensure_session(self, count = 0, *args, **kwargs):
    if(not self.session_expired()):
      return True
    
    if count > 5:
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = "NOT AUTHENTICATED",
        message = f"Error waiting for authentication"
      )
    
    if(self.is_authenticating_session()):
      poll2(
        lambda: self.is_authenticating_session(),
        ignore_exceptions=(Exception,),
        timeout=self.get_aws_poll_authenticate(),
        step=0.1
      )
      return self.ensure_session(count= count+1)
      
    self._set_authenticating_session(is_authenticating_session= True)

    try:
      self.authenticate_session()
    finally:
      self._set_authenticating_session(is_authenticating_session= False)
    
    return self.ensure_session()
  
  def is_authenticating_session(self, *args, **kwargs):
    if(hasattr(self, "_is_authenticating_session")):
      return self._is_authenticating_session
    
    return False
  
  def _set_authenticating_session(self, is_authenticating_session, *args, **kwargs):
    if not is_authenticating_session:
      self._get_assumed_role_credentials(unset= True)
      self.session_expired(refresh= True)
      
    self._is_authenticating_session = is_authenticating_session
    return self.is_authenticating_session()

  def _get_organization_client(self, *args, **kwargs):   
    if hasattr(self, "_org_client"):
      return self._org_client
    
    self._org_client = self.get_boto_client(
      client= 'organizations', 
      account=self.get_organization_account(),
      role = None,
      region = None
    )
    return self._get_organization_client()

  
  def get_boto_config(self, region= None, max_attempts = 10, read_timeout = 900, connect_timeout = 10, max_pool_connections = 20,   *argv, **kwargs):
    config = botocore_config_config(
      retries = {
      'max_attempts': max_attempts,
      'mode': 'standard'
      },
      read_timeout = read_timeout,
      connect_timeout = connect_timeout,
      max_pool_connections = max_pool_connections
    )

    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= region):
      return config
    
    config.region_name = region
    return config

  def get_boto_client(self, client, account=None, role = None, region = None, *argv, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= region):
      region = self.get_default_region()
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= role):
      role = self.get_default_rolename()

    if account is None or (self.get_common().helper_type().general().is_type(obj= account, type_check= str) and self.get_common().helper_type().string().is_null_or_whitespace(string_value= account)):
      account = self.get_default_account()

    cache_key = self._get_boto_client_key(client= client, account= account, role= role, region= region)
    if self._get_created_boto_clients().get(cache_key) is not None:
      return  self._get_created_boto_clients().get(cache_key)

    session = self.get_boto_session(
        account=account,
        region=region,
        role=role
      )


    if hasattr(session, "create_client"):
       self._get_created_boto_clients()[cache_key] = session.create_client(client, config= self.get_boto_config(region= region, *argv, **kwargs))
       return self._get_created_boto_clients()[cache_key]
       
    self._get_created_boto_clients()[cache_key] = session.client(client, config= self.get_boto_config(region= region, *argv, **kwargs))
    return self._get_created_boto_clients()[cache_key]

  def _auto_parse_aws_expiration(self, expiration, *argv, **kwargs):
    if self.get_common().helper_type().general().is_type(obj= expiration, type_check= int):
      return self.get_common().helper_type().datetime().get_from_timestamp(time_delta=expiration)
    
    if self.get_common().helper_type().general().is_type(obj= expiration, type_check= str):
      return self.get_common().helper_type().datetime().parse_iso(iso_datetime_str=expiration)
    

  def convert_assume_role_credentials_export(self, credentials):
    
    outputData = f"\nexport AWS_ACCESS_KEY_ID=\"{credentials['AccessKeyId']}\"\n"
    outputData += f"export AWS_SECRET_ACCESS_KEY=\"{credentials['SecretAccessKey']}\"\n"
    outputData += f"export AWS_SESSION_TOKEN=\"{credentials['SessionToken']}\"\n"
    outputData += f"export AWS_DEFAULT_REGION=\"{self.get_default_region()}\"\n"

    return outputData
  
  def convert_assume_role_credentials_cli(self, credentials):
    # https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-sourcing-external.html
    expiration = self._auto_parse_aws_expiration(
      expiration= credentials["expiration"]
    )

    return {
      "Version": 1,
      "AccessKeyId": credentials["accessKeyId"],
      "SecretAccessKey": credentials["secretAccessKey"],
      "SessionToken": credentials["sessionToken"],
      "Expiration": self.get_common().helper_type().datetime().get_iso_datetime(
        dt= expiration
      )
    }
        
  def convert_assume_role_credentials_boto_session(self, credentials):
    expiration = self._auto_parse_aws_expiration(
      expiration= credentials["expiration"]
    )
    

    return {
      "access_key": credentials["accessKeyId"],
      "secret_key": credentials["secretAccessKey"],
      "token": credentials["sessionToken"],
      "expiry_time": self.get_common().helper_type().datetime().get_iso_datetime(
        dt= expiration
      )
    }


  def __get_boto_session(self, role = None, region = None, profile = None, **kwargs):
    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile):
      return boto_session(profile_name= profile) if self.get_common().helper_type().string().is_null_or_whitespace(string_value= region) else boto_session(profile_name= profile, region_name= region)

    return botocore_session.get_session()

  def get_boto_session(self, account=None, role = None, region = None, profile = None, *args, **kwargs):
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= role):
      role = self.get_default_rolename()
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= region):
      region = self.get_default_region()

    if account is None or (self.get_common().helper_type().general().is_type(obj= account, type_check= str) and self.get_common().helper_type().string().is_null_or_whitespace(string_value= account)):
      account = self.get_default_account()

    cache_key = self._get_boto_session_key(account= account, role= role, region= region, profile= profile)
    if self._get_created_boto_sessions().get(cache_key) is not None:
      return self._get_created_boto_sessions().get(cache_key)

    session = self.__get_boto_session(
      account=account, role = role, region = region, profile = profile
    )     

    credentials = botocore_credentials.RefreshableCredentials.create_from_metadata(
      metadata=self.convert_assume_role_credentials_boto_session(self.assume_role(account=account, role=role)),
      refresh_using=lambda: self.convert_assume_role_credentials_boto_session(self.assume_role(account=account, role=role, refresh= True)),
      method="sts-assume-role",
    )

    session._credentials = credentials 
    session.set_config_variable("region", region)
      
    self._get_created_boto_sessions()[cache_key] = boto_session(botocore_session= session)
    return self.get_boto_session(account=account, role = role, region = region, profile = profile, *args, **kwargs)
  
  def _get_account_regions_costexplorer(self, account, services = None, *args, **kwargs ):
    
    ce_client = self.get_boto_client(
      client= "ce",
      account=account
    )

    filter = {}
    granularity = "MONTHLY"
    
    utc_now = self.get_common().helper_type().datetime().get()
    time_period = {
      "Start":(self.get_common().helper_type().datetime().datetime_as_string(
        dt= (utc_now + self.get_common().helper_type().datetime().time_delta(months= -1, dt= utc_now)),
        dt_format= "%Y-%m-%d"
      )),
      "End": (self.get_common().helper_type().datetime().datetime_as_string(
        dt= utc_now,
        dt_format= "%Y-%m-%d"
      ))
    }   
    metrics = ["NETUNBLENDEDCOST"]
    group_by= [{"Type":"DIMENSION", "Key":"REGION"}]

    get_cost_usage_params = {
      "Granularity":granularity,
      "TimePeriod": time_period,
      "Metrics": metrics,
      "GroupBy": group_by
    }

    if not services is None:
      filter["Dimensions"] = {
        "Key": "SERVICE",
        "Values": services,
        "MatchOptions": ["EQUALS", "CASE_SENSITIVE"]
      }
      get_cost_usage_params["Filter"] = filter


    try:
      discovered_resultsbytime = self.general_boto_call_array(
        boto_call = lambda item: ce_client.get_cost_and_usage(**item),
        boto_params = get_cost_usage_params,
        boto_nextkey = "NextToken",
        boto_key="ResultsByTime"
      )

      regions = {}

      for itemTime in discovered_resultsbytime:
        for itemGroup in itemTime["Groups"]:
          for item in itemGroup["Keys"]:
            item_key = self.get_common().helper_type().string().set_case(string_value= item, case= "lower")
            if item_key == "noregion":
              continue
            if not item_key in regions:
              regions[item_key] = item
      return regions
    
    except Exception as err:
      self.get_common().get_logger().exception(msg=f"Could not pull regions dynamically: {err}", extra= {"exception": err})
      return {
        self.get_common().helper_type().string().set_case(string_value= region, case= "lower"):region for region in self.default_region_resources
      }
  
  def get_accounts_regions_costexplorer(self, accounts, services, role= None):     
    if role is None:
      role = self.get_default_rolename()
    regions = { }
    
    for account in accounts:      
      regions[self.get_account_id(account)] = self.get_account_regions_costexplorer(
        account=account,
        role=role,
        services= services
      )

    return regions

  def get_account_regions_costexplorer(self, account, role, services):     
    # to get a list of all services for an account
    # aws ce get-dimension-values --time-period "Start=2022-09-01,End=2022-09-11" --dimension SERVICE

    return self._get_account_regions_costexplorer(
      account= account, 
      role= role,
      services= services
    )
  
  def _get_regions_for_resource_configurationaggregator(self, account, role, configuration_accountId, configuration_aggregatorName, configuration_aggregatorRegion, configuration_aggregator_recourcetype):
    
    config_client = self.get_boto_client(
      client= "config",
      account= self.make_account(Id= configuration_accountId),
      role= role,
      region= configuration_aggregatorRegion)

    retryCount = 10
    currentAttempt = 0

    filter = {}
    filter["AccountId"] = self.get_account_id(account)    
    filter["ResourceType"] =configuration_aggregator_recourcetype


    while currentAttempt < retryCount:
      currentAttempt+=1

      try:
        discovered_resources = config_client.get_aggregate_discovered_resource_counts(
          ConfigurationAggregatorName = configuration_aggregatorName,
          GroupByKey = "AWS_REGION",
          Filters=filter
          
        )
        break
      except ClientError as err:
        if currentAttempt > 2:
          self.get_common().get_logger().exception(msg=f"_get_regions_for_resource: {err}", extra= {"exception": err})
        discovered_resources = None
        sleepTime = (2**currentAttempt)+randint(1,10)
        if sleepTime > 30:
          sleepTime = 30
        time.sleep(sleepTime)

    if discovered_resources is None:
      return { }

    regions = {self.get_account_id(account):[] }
    for item in discovered_resources["GroupedResourceCounts"]:

      if item["GroupName"] in regions[self.get_account_id(account)]:
        continue

      regions[self.get_account_id(account)].append(item["GroupName"])
  
    while "NextToken" in discovered_resources:
      currentAttempt = 0
      nextToken = discovered_resources["NextToken"]

      while currentAttempt < retryCount:
        currentAttempt+=1  
        try:  
          discovered_resources = config_client.get_aggregate_discovered_resource_counts(
            ConfigurationAggregatorName = configuration_aggregatorName,
            GroupByKey = "AWS_REGION",
            Filters=filter,
            NextToken= nextToken
          )            
          break
            
        except ClientError as err:
          if currentAttempt > 2:
            self.get_common().get_logger().exception(msg=f"_get_regions_for_resource - next loop: {err}", extra= {"exception": err})
            print("_get_regions_for_resource - next loop: {}".format(err))
          sleepTime = (2**currentAttempt)+randint(1,10)
          if sleepTime > 30:
            sleepTime = 30
          time.sleep(sleepTime)        
      
      for item in discovered_resources["GroupedResourceCounts"]:
        if item["GroupName"] in regions[self.get_account_id(account)]:
          continue

        regions[self.get_account_id(account)].append(item["GroupName"])


    return regions
  
  def get_regions_account_configurationaggregator(self, accounts, role, configuration_accountId, configuration_aggregatorName, configuration_aggregatorRegion, configuration_aggregator_recourcetype):
    regions = { }

    for account in accounts:
      if self.verbose:
        if "accountName" in account:
          print("Getting Regions for Account: {} ({})".format(self.get_account_name(account),self.get_account_id(account)))
        else:
          print("Getting Regions for Account: {}".format(self.get_account_id(account)))
      for resource_type in configuration_aggregator_recourcetype:
        tmpregions = self._get_regions_for_resource_configurationaggregator(account, role, configuration_accountId, configuration_aggregatorName, configuration_aggregatorRegion, resource_type)
        for item in tmpregions:
          if item in regions:
            regions[item] = list(set(regions[item] + tmpregions[item]))    
            continue

          regions[item] = tmpregions[item]

    return regions
  
  def _get_accounts(self, refresh = False, include_suspended = False):
    
    if hasattr(self, "_account_list") and not refresh:
      return_list = "active" if not include_suspended else "all"
      if self._account_list is not None and len(self._account_list) > 0:
        if return_list in self._account_list:
          return self._account_list[return_list]
    
    
    self._account_list = {
      "all": self.general_boto_call_array(
        boto_call=lambda item: self._get_organization_client().list_accounts(**item),
        boto_params={},
        boto_nextkey = "NextToken",
        boto_key="Accounts"
      )
    }
    
    self._account_list["active"] = [ acct for acct in self._account_list["all"] if self.get_common().helper_type().string().set_case(string_value= acct["Status"], case= "lower") != "suspended" ]
    return self._get_accounts(refresh= False, include_suspended= include_suspended) 

  def get_accountids_by_ou(self, org_ou, exclude_ous = None, **kwargs):
    if org_ou is None:
      org_ou = []
    
    if exclude_ous is None:
      exclude_ous = []
    
    if self.get_common().helper_type().general().is_type(obj= org_ou, type_check= str) and not self.get_common().helper_type().string().is_null_or_whitespace(string_value= org_ou): 
      org_ou = [ ou.strip() for ou in self.get_common().helper_type().string().split(string_value= org_ou) if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= ou) ]
    
    account_list = []
    for ou in org_ou:
      child_ous = [child_ou["Id"] for child_ou in self.general_boto_call_array(
        boto_call=lambda item: self._get_organization_client().list_children(**item),
        boto_params={"ParentId": ou, "ChildType": "ORGANIZATIONAL_UNIT"},
        boto_nextkey = "NextToken",
        boto_key="Children"
      ) if f'-{self.get_common().helper_type().string().set_case(string_value= child_ou["Id"], case= "lower")}' not in exclude_ous]
      account_list += self.get_accountids_by_ou(org_ou= child_ous, exclude_ous= exclude_ous)
      account_list += [account["Id"] for account in self.general_boto_call_array(
        boto_call=lambda item: self._get_organization_client().list_children(**item),
        boto_params={"ParentId": ou, "ChildType": "ACCOUNT"},
        boto_nextkey = "NextToken",
        boto_key="Children"
      )]
    
    return list(dict.fromkeys(account_list))
  
  def get_accounts(self, account = None, refresh = False, include_suspended = False):    
    all_accounts = self._get_accounts(refresh=refresh, include_suspended=include_suspended)
    if account is None:
      return all_accounts
    
    if self.get_common().helper_type().general().is_type(obj= account, type_check= str) and not self.get_common().helper_type().string().is_null_or_whitespace(string_value= account):
      account = [ self.get_account_id(account= acct) for acct in self.get_common().helper_type().string().split(string_value= account) if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= acct) ]

    if not self.get_common().helper_type().general().is_type(obj= account, type_check= list):
      self.get_common().get_logger().warning(f'unknown data type for accounts {type(account)}, when trying to get accounts')
      return all_accounts

    search_accounts = [ self.get_account_id(account= acct) for acct in account if not self.get_account_id(account= acct) .startswith("ou-") and not self.get_account_id(account= acct) .startswith("-") ]
    search_accounts_ous = [ self.get_account_id(account= acct)  for acct in account if self.get_account_id(account= acct) .startswith("ou-") ]
    exclude_accounts_ous = [ self.get_account_id(account= acct)  for acct in account if self.get_account_id(account= acct) .startswith("-ou-") ]
    exclude_accounts = [ self.get_account_id(account= acct)  for acct in account if self.get_account_id(account= acct).startswith("-") and not self.get_account_id(account= acct).startswith("-ou-") ]
    account = list(dict.fromkeys(search_accounts + self.get_accountids_by_ou(org_ou= search_accounts_ous, exclude_ous= exclude_accounts_ous)))

    return [ acct for acct in all_accounts if f'-{self.get_account_id(account= acct) }' not in exclude_accounts and  (len(account) < 1 or self.get_common().helper_type().list().find_item(data= account, filter= lambda item: self.get_account_id(account= item)  == self.get_account_id(account= acct)) is not None) ]
  
  def get_resource_groups(self, account, region, filters_rg = [], rg_client = None):
    if rg_client is None:
      rg_client = self.get_boto_client(
        client= 'resource-groups', 
        account= account,
        role = None,
        region = region
      )
    
    # Global does not seem like a valid option
    if (self.get_common().helper_type().string().set_case(string_value= region, case= "lower") == "global"):
        return []

    boto_params = {}
    if len(filters_rg) > 0:
      boto_params["Filters"] = filters_rg

    return self.general_boto_call_array(
      boto_call=lambda item: rg_client.list_groups(**item),
      boto_params= boto_params,
      boto_nextkey = "NextToken",
      boto_key="GroupIdentifiers"
    )
  
  def get_resource_group_from_resource(self, account, region, filters_rg = [], filters_resource = [], rg_client = None, resource_groups = None, treat_badfilter_err_as_empty = True):
    if rg_client is None:
      rg_client = self.get_boto_client(
        client= 'resource-groups', 
        account= account,
        role = None,
        region = region
      )
      
    if resource_groups is None or (resource_groups is not None and len(resource_groups)<1):
      resource_groups = self.get_resource_groups(account=account, region= region,filters=filters_rg, rg_client= rg_client)

      if resource_groups is None or (resource_groups is not None and len(resource_groups)<1):
        return {}

    resources = {} 
    boto_params = {}
    if len(filters_resource) > 0:
      boto_params["Filters"] = filters_resource
    
    boto_nextkey = "NextToken"
    for rg in resource_groups:
      boto_params["Group"] = rg["GroupArn"]
      if boto_nextkey in boto_params:
        boto_params.pop(boto_nextkey)
      try:
        for resource in self.general_boto_call_array(
            boto_call=lambda item: rg_client.list_group_resources(**item),
            boto_params= boto_params,
            boto_nextkey = boto_nextkey,
            boto_key="Resources",
            error_codes_raise = ["BadRequestException"]
          ):
        
          if resource.get("Identifier").get("ResourceArn").lower() not in resources:
            resources[resource.get("Identifier").get("ResourceArn").lower()] = []
          
          resources[resource.get("Identifier").get("ResourceArn").lower()].append(rg["GroupName"])
      
      except Exception as err:
        if treat_badfilter_err_as_empty and "filters not valid" in str(err).lower():
          continue

        raise err

    return resources
  
  def get_resource_tags_as_dictionary(self, resource, *args, **kwargs):
    if resource is None:
      return None

    tags = None
    if not self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      if hasattr(resource, "extra_tags"):
        if resource.extra_tags is not None and len(resource.extra_tags) > 0:
          return resource.extra_tags
      if hasattr(resource, "tags"):
        tags = resource.tags
        
    elif self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      if resource.get("extra_tags") is not None and len(resource.get("extra_tags")) > 0:
        return resource.get("extra_tags")
      tags = resource.get("tags")

    if tags is None:
      tags = {}
    
    if self.get_common().helper_type().general().is_type(obj= tags, type_check= list ):
        tags = {tag["Key"]:tag["Value"] for tag in tags}

    return tags


from threemystic_cloud_client.cloud_providers.aws.action_test.base_class.base import cloud_client_aws_test_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers
from threemystic_cloud_client.cloud_providers.aws.action_test.step_2 import cloud_client_aws_test_step_2 as nextstep
from threemystic_common.base_class.base_script_options import base_process_options
import textwrap, argparse

class cloud_client_aws_test_step_1(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_aws_test", *args, **kwargs)

    self._process_options = base_process_options(common= self.get_common())
    self._token_parser_args = {   }

    self._process_cli_args()
  
  def _process_cli_args(self, *args, **kwargs):
    parser = self._process_options.get_parser(
      parser_init_kwargs = {
        "prog": "3mystic_cloud_client --test -p aws",
        "formatter_class": argparse.RawDescriptionHelpFormatter,
        "description": textwrap.dedent('''\
        Tests the profile configuration
        '''),
        "add_help": False,
        "epilog": ""
      },
      parser_args = self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        self._token_parser_args,
        {
          "--profile": {
            "default": None, 
            "type": str,
            "dest": "test_profile",
            "help": "The 3Mystic profile to use.",
            "action": 'store'
          }
        }
      ])
    )


    processed_arg_info = self._process_options.process_opts(
      parser = parser
    )

    self._processed_arg_info = processed_arg_info.get("processed_data")

  def step(self, *args, **kwargs):
    if not super().step( *args, **kwargs):
      return
    
    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._processed_arg_info.get("test_profile")):
      if self.config_profile_name_exists(profile_name= self._processed_arg_info.get("test_profile")):
        return nextstep(init_object = self).step( profile_name= self._processed_arg_info.get("test_profile"))


    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "profile": {
            "validation": lambda item: self.config_profile_name_exists( profile_name= item),
            "messages":{
              "validation": f"Please enter a valid existing Cloud Client Profile",
            },
            "conversion": lambda item: self.get_common().helper_type().string().set_case(string_value= self.get_common().helper_type().string().trim(string_value= item), case= "lower"),
            "desc": f"What Cloud Client Profile to load",
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": True,
            "default": self.get_default_profile_name()
        }
      }
    )

    if response is None:
      return

    if not self.config_profile_name_exists(profile_name= response["profile"].get("formated")):
      print(f"Profile Not Found: {response['profile'].get('formated')}")
      return

    
    nextstep(init_object = self).step( profile_name= response['profile'].get('formated'))
    
  



from threemystic_cloud_client.cloud_providers.azure.action_test.base_class.base import cloud_client_azure_test_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers
from threemystic_common.base_class.base_script_options import base_process_options
import textwrap, argparse


class cloud_client_azure_token_step_1(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_azure_token", *args, **kwargs)

    self._process_cli_args()
    
  def _process_cli_args(self, *args, **kwargs):
    process_options = base_process_options(common= self.get_common())
    token_parser_args = {   }
    self._arg_parser = process_options.get_parser(
      parser_init_kwargs = {
        "prog": "3mystic_cloud_client --token -p azure",
        "formatter_class": argparse.RawDescriptionHelpFormatter,
        "description": textwrap.dedent('''\
        Requires additional settings.
          --tenant is required
          --resource is required

          To learn more please see: https://learn.microsoft.com/en-us/cli/azure/account?view=azure-cli-latest#az-account-get-access-token
        '''),
        "add_help": False,
        "epilog": ""
      },
      parser_args = self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        token_parser_args,
        {
          "--resource": {
            "default": None, 
            "type": str,
            "dest": "token_resource",
            "help": "Azure resource endpoints in AAD v1.0.",
            "action": 'store'
          },
          # "--resource-type": {
          #   "default": None, 
          #   "type": str,
          #   "dest": "token_resource_type",
          #   "help": "Type of well-known resource. ie aad-graph, arm, ms-graph",
          #   "action": 'store'
          # },
          # "--scope": {
          #   "default": None, 
          #   "type": str,
          #   "dest": "token_scope",
          #   "help": "Comma Seperated AAD scopes in AAD v2.0.",
          #   "action": 'store'
          # },
          "--tenant": {
            "default": None, 
            "type": str,
            "dest": "token_tenant",
            "help": "Tenant ID for which the token is acquired.",
            "action": 'store'
          }
        }
      ])
    )


    processed_info = process_options.process_opts(
      parser = self._arg_parser
    )

    self._processed_arg_info = processed_info.get("processed_data")
  
    
  def step(self, *args, **kwargs):
    if not super().step( *args, **kwargs):
      return
    
    if (self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._processed_arg_info.get("token_tenant")) or (self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._processed_arg_info.get("token_resource")) and 
      self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._processed_arg_info.get("token_resource_type")))):
      self._arg_parser.print_help()
      return

    from threemystic_cloud_client.cloud_client import cloud_client
    azure_client = cloud_client(logger= self.get_common().get_logger(), common=self.get_common()).client(
      provider= "azure"
    )
    
    tenant_credential = azure_client.get_tenant_credential(
      tenant= self._processed_arg_info.get("token_tenant")
    )
    
    if not (self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._processed_arg_info.get("token_resource"))): 
      token_data = tenant_credential.get_token(self._processed_arg_info.get("token_resource"))
      print(self.get_common().helper_json().dumps(data= {
        "token": token_data.token,
        "expires_on": token_data.expires_on
        }))
    
  



import sys
from threemystic_common.base_class.base_script_options import base_process_options
from threemystic_cloud_client.cloud_client import cloud_client as threemystic_client
from threemystic_cloud_client.cli.actions.config import cloud_client_config as user_action_config

class cloud_client_cli(base_process_options):
  def __init__(self, cloud_client = None, *args, **kwargs):    
    self._cloud_client = threemystic_client() if cloud_client is None else cloud_client
    
    super().__init__(common= self._cloud_client.get_common(), *args, **kwargs)

    self.parser = self.get_parser(
      parser_init_kwargs = {
        "description": "One Action is required",
        "add_help": False,
      },
      parser_args = {
        "--help,-h": {
          "default": False,
          "dest": "client_help",
          "help": "Display Help",
          "action": 'store_true'
        },
        # I can create other actions just by duplication this and changing the const,
        "--version": {
            "default": None, 
            "const": "version",
            "dest": "client_action",
            "help": "Action: outputs the versions of the app being used.",
            "action": 'store_const'
        },
        "--config,-c": {
            "default": None, 
            "const": "config",
            "dest": "client_action",
            "help": "Action: This is so you can setup the cloud client to work with various providers",
            "action": 'store_const'
        },
        "--test,-t": {
            "default": None, 
            "const": "test",
            "dest": "client_action",
            "help": "Action: This is so you can test the config setup to ensure the base connection is good",
            "action": 'store_const'
        },
        "--token": {
            "default": None, 
            "const": "token",
            "dest": "client_action",
            "help": "Action: This is so that you can generate the required token.",
            "action": 'store_const'
        },
        "--generate,-g": {
            "default": None, 
            "const": "generate",
            "dest": "client_action",
            "help": "Action: For providers like aws it is easier to have a profile when interacting with the accounts. This will help generate the various profiles.",
            "action": 'store_const'
        },
        "--provider,-p": {
            "default": None, 
            "type": str,
            "choices": self._cloud_client.get_supported_providers(),
            "dest": "client_provider",
            "help": "Provider: This is to set the provider that should be used",
            "action": 'store'
        }
      }
    )

    processed_info = self.process_opts(
      parser = self.parser
    )

    for key, value in processed_info["processed_data"].items():
      setattr(self, f"_{key}", value)
    
    
  def process_client_action(self, force_action = None, *args, **kwargs):
    if self._cloud_client.get_common().helper_type().string().is_null_or_whitespace(string_value= force_action):
      force_action = self.__get_client_acount()

    if self._cloud_client.get_common().helper_type().string().is_null_or_whitespace(string_value= force_action):
      self.parser.print_help()
      return
    
    if force_action == "version":
      self.version_dispaly()
      return
    
    if self._cloud_client.get_common().helper_type().string().is_null_or_whitespace(string_value=  self.__get_client_provider()):
      self.parser.print_usage()

    if force_action == "config":
      user_action_config(cloud_client= self._cloud_client).main(provider= self.__get_client_provider())
      return

    if force_action == "test":
      from threemystic_cloud_client.cli.actions.action_test import cloud_client_test as user_action
      user_action(cloud_client= self._cloud_client).main(provider= self.__get_client_provider())
      return

    if force_action == "token":
      from threemystic_cloud_client.cli.actions.action_token import cloud_client_token as user_action
      user_action(cloud_client= self._cloud_client).main(provider= self.__get_client_provider())
      return

    if force_action == "generate":
      from threemystic_cloud_client.cli.actions.action_generate import cloud_client_generate as user_action
      user_action(cloud_client= self._cloud_client).main(provider= self.__get_client_provider())
      return

    return

  def version_dispaly(self, *args, **kwargs): 
    print(f"You currenly have installed")
    print(f"3mystic_cloud_client: v{self._cloud_client.version()}")
    print(f"3mystic_common: v{self._cloud_client.get_common().version()}")
    print()
    print(f"Current supported cloud providers: {self._cloud_client.get_supported_providers()}")
    print(f"Cloud Providers config status: ")
    for cloud_provider in self._cloud_client.get_supported_providers():
      print(f"{cloud_provider}:  {user_action_config(cloud_client= self._cloud_client).provider_config_status(provider= cloud_provider)}")
    

  def __get_client_provider(self, *args, **kwargs):
    if not hasattr(self, "_client_provider"):
      return None
    
    return self._client_provider

  def __get_client_acount(self, *args, **kwargs):
    if not hasattr(self, "_client_action"):
      return None
    
    return self._client_action
  def main(self, *args, **kwargs):    
    if self.__get_client_acount() is None:
      print(f"Thank you for using the 3 Mystic Apes Cloud Client.")
      self.version_dispaly()
      print()
      self.parser.print_help()
      return
    
    self.process_client_action( )

def main(*args, **kwargs):    
  cloud_client_cli(*args, **kwargs).main(*args, **kwargs)
    

if __name__ == '__main__':   
  cloud_client_cli().main(sys.argv[1:])


from base_class.base import base

class client_aws_sso(base): 
  """This is a library to help with the interaction with the cloud providers"""


  def __init__(self, *args, **kwargs) -> None: 
    super().__init__(*args, **kwargs)
  
  


from threemystic_cloud_client.cloud_providers.aws.action_test.base_class.base import cloud_client_aws_test_base as base
from threemystic_common.base_class.base_script_options import base_process_options
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers
import textwrap, argparse
from threemystic_cloud_client.cloud_client import cloud_client
import re

class cloud_client_aws_generate_step_1(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_aws_token", *args, **kwargs)
    
    
    self._process_cli_args()
    self.__generate_format_options = {
      "accountId": {
        "description": "This will be replaced with the account id",
        "handler": lambda item: self._aws_client.get_account_id(account= item["account"])
      },
      "accountName": {
        "description": "This will be replaced with the account name spaces will be replaced with dashes",
        "handler": lambda item: self.get_common().helper_type().string().join("-", str_array= self.get_common().helper_type().string().split(
          string_value= self._aws_client.get_account_name(account= item["account"]),
          separator= "[^a-zA-Z0-9\-\_]"
        ))
      },
      "roleName": {
        "description": "The user role from the aws connection.",
        "handler": lambda item: self._aws_client.get_default_rolename()
      },
      "profileName": {
        "description": "The 3Mystic Cloud Client Profile Name",
        "handler": lambda item: self._aws_client.get_profile().get("profile_name")
      }
    }
    
  def _process_cli_args(self, *args, **kwargs):
    process_options = base_process_options(common= self.get_common())
    token_parser_args = { 
      "--help,-h": {
        "default": False,
        "dest": "client_help",
        "help": "Display Help",
        "action": 'store_true'
      }
    }
    self._arg_parser = process_options.get_parser(
      parser_init_kwargs = {
        "prog": "3mystic_cloud_client --generate -p aws",
        "formatter_class": argparse.RawDescriptionHelpFormatter,
        "description": textwrap.dedent('''\
        This tool will help auto generate the aws config file. It will override anyhting alreayd there and will not generate a default profile.
        '''),
        "add_help": False,
        "epilog": ""
      },
      parser_args = self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        token_parser_args,
        {
          "--profile": {
            "default": None, 
            "type": str,
            "dest": "token_profile",
            "help": "The 3Mystic AWS Profile to use. If not provided the default will be used",
            "action": 'store'
          },
        },
      ])
    )


    processed_info = process_options.process_opts(
      parser = self._arg_parser
    )
    
    self._processed_arg_info = processed_info.get("processed_data")
  
  def get_valid_profile_name_options(self, *args, **kwargs):
    return self.__generate_format_options
  
  def get_valid_profile_name_options_display(self, *args, **kwargs):
    for key, item in self.get_valid_profile_name_options().items():
      print(f'{key}: {item.get("description")}')
  
  def step(self, *args, **kwargs):
    if not super().step( *args, **kwargs):
      return
    
    if (self.get_common().helper_type().bool().is_true(check_value= self._processed_arg_info.get("client_help"))):
      self._arg_parser.print_help()
      return
    
    self._aws_client = cloud_client(logger= self.get_common().get_logger(), common=self.get_common()).client(
      provider= "aws",
      profile_name= self._processed_arg_info.get("token_profile")
    )
    
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "format": {
            "validation": lambda item: ((not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item)) 
                                        and self.get_common().helper_type().string().set_case(string_value= self.get_common().helper_type().string().trim(string_value= item), case= "lower") not in ["default", "profile"]
                                        ),
            "messages":{
              "validation": f"The processed name cannot be profile or default.",
            },
            "conversion": lambda item: self.get_common().helper_type().string().trim(string_value= item),
            "desc": f"What should the format be?\nFormat Variables:\n{self.get_valid_profile_name_options_display()}",
            "default": "{accountName}-{roleName}",
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": True
        }
      }
    )

    if response is None:
      return
    
    self.step_process_generation(format= response.get("format").get("formated"))

  def step_process_generation(self, format, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= format):
      format = "{accountName}-{roleName}"

    aws_config = self.get_common().helper_config().load(
      config_type = "config",
      path= self.get_aws_user_path_config()
    )
    
    for account in self._aws_client.get_accounts():
      profile = format
      for key, item in self.get_valid_profile_name_options().items():
        profile= re.sub(
          pattern=f"{{{key}}}", 
          repl= item.get("handler")({"account": account}),
          string= profile, count=0)
      
      aws_config[f"profile {profile}"] = {
        "credential_process": f"3mystic_cloud_client --token -p aws --format cli --profile {self._aws_client.get_profile().get('profile_name')} --account {self._aws_client.get_account_id(account= account)}",
        "region": self._aws_client.get_default_region()
      }
      print(f'Profile - {profile} - Generated for {self._aws_client.get_account_name(account= account)} - {self._aws_client.get_account_id(account= account)}')
  
    with open(self.get_aws_user_path_config(), 'w') as configfile:
      aws_config.write(configfile)

    print("-----------------------------")
    print()
    print()
    print("AWS Config Updated")
    print()
    print()
    print("-----------------------------")
      

      

    
  



from threemystic_common.base_class.base_provider import base


class cloud_client(base): 
  """This is a library to help with the interaction with the cloud providers"""

  def __init__(self, logger = None, common = None, *args, **kwargs) -> None: 
    super().__init__(provider= "", common= common, logger_name= "cloud_client", logger= logger, *args, **kwargs)
    
  def version(self, *args, **kwargs):
    if hasattr(self, "_version"):
      return self._version
    import threemystic_cloud_client.__version__ as __version__
    self._version = __version__.__version__
    return self.version()
    
  def get_supported_providers(self, *args, **kwargs):
    return super().get_supported_providers()

  def __set_provider_azure(self, provider, *args, **kwargs):
    from threemystic_cloud_client.cloud_providers.azure.client.auto_client import cloud_client_azure_client_auto as client
    self._client[provider] = client(
      common= self.get_common(), *args, **kwargs
    ).get_client()
  
  def __set_provider_aws(self, provider, *args, **kwargs):
    from threemystic_cloud_client.cloud_providers.aws.client.auto_client import cloud_client_aws_client_auto as client
    self._client[provider] = client(
      common= self.get_common(), *args, **kwargs
    ).get_client()

  def init_client(self, provider, *args, **kwargs):
    provider = self.get_common().helper_type().string().set_case(string_value= provider, case= "lower") if provider is not None else ""

    if provider not in self.get_supported_providers():
      raise self.get_common().exception().exception(
        exception_type = "argument"
      ).not_implemented(
        logger = self.get_common().get_logger(),
        name = "provider",
        message = f"Unknown Cloud Provided: {provider}.\nSupported Cloud Providers{self.get_supported_providers()}"
      )

    if not hasattr(self, "_client"):
      self._client = {}

    if self._client.get(provider) is not None:
      return

    if provider == "azure":
      self.__set_provider_azure(provider= provider, *args, **kwargs)
      return
    
    if provider == "aws":
      self.__set_provider_aws(provider= provider, *args, **kwargs)
      return  
       
    raise self.get_common().exception().exception(
      exception_type = "argument"
    ).not_implemented(
      logger = self.get_common().get_logger(),
      name = "provider",
      message = f"Unknown Cloud Provided: {provider}.\nSupported Cloud Providers{self.get_supported_providers()}"
    )

  def client(self, provider, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= provider):
      raise self.get_common().exception().exception(
        exception_type = "argument"
      ).not_implemented(
        logger = self.get_common().get_logger(),
        name = "provider",
        message = f"provider cannot be null or whitespace"
      )
  
    provider = self.get_common().helper_type().string().set_case(string_value= provider, case= "lower")
    if not hasattr(self, "_client"):
      self.init_client(provider= provider,  *args, **kwargs)
      return self.client(provider= provider, *args, **kwargs)
    
    return self._client.get(provider)


from threemystic_cloud_client.cloud_providers.azure.client.base_class.base import cloud_client_azure_client_base as base
from azure.identity import AzureCliCredential



class cloud_client_azure_client_cli(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_azure_client_sso", *args, **kwargs)

  
  def _login(self, on_login_function = None, tenant = None, *args, **kwargs):
 
    tenant_id = f' --tenant {self.get_tenant_id(tenant= tenant)}' if tenant is not None else ""

    return self._az_cli(
      command= f"az login{tenant_id} --allow-no-subscriptions",
      on_login_function = on_login_function
    )
        
  def _get_tenant_credential(self, tenant, *args, **kwargs):
    return AzureCliCredential(tenant_id= self.get_tenant_id(tenant= tenant))

  


from threemystic_cloud_client.cli.actions.base_class.base import cloud_client_action_base as base
from threemystic_common.base_class.base_script_options import base_process_options
import textwrap, argparse

class cloud_client_token(base):
  def __init__(self, *args, **kwargs):
    super().__init__(action= "--token", *args, **kwargs)    


  def _process_provider_aws(self, *args, **kwargs):    
    from threemystic_cloud_client.cloud_providers.aws import cloud_client_aws as client
    client(common= self._cloud_client.get_common()).action_token()


  def _process_provider_azure(self, *args, **kwargs):
    
    from threemystic_cloud_client.cloud_providers.azure import cloud_client_azure as client
    client(common= self._cloud_client.get_common()).action_token()      

      

  



from threemystic_cloud_client.cloud_providers.aws.base_class.base import cloud_client_provider_aws_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers

class cloud_client_aws_config_base(base):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def update_config_profile(self, profile_name, profile_data, auto_save = True, *args, **kwargs):

    profile_name = self.get_common().helper_type().string().set_case(string_value= profile_name, case= "lower")
    self.get_config_profiles()[profile_name] = profile_data
    if(profile_data["default_profile"]):
      if(profile_name != self.get_default_profile_name() and self.has_default_profile()):

        for existing_profile, existing_profile_data in self.get_config_profiles().items():
          if(not existing_profile_data["default_profile"] or existing_profile == profile_name):
            continue

          existing_profile_data["default_profile"] = False
    
    if auto_save:
      self._save_config()
  
  def update_is_cli_installed(self, is_cli_installed, *args, **kwargs):
    config = self.get_config()
    
    config["cli_installed"] = is_cli_installed

    self._save_config()
  
  def update_sdk_auth(self, sdk_auth = "cli", *args, **kwargs):
    config = self.get_config()
    
    config["sdk_auth"] = sdk_auth

    self._save_config()
     
  def step(self, force_cli_installed_prompt = False, *args, **kwargs):
    
    if (self.is_cli_installed() != True or force_cli_installed_prompt):   
      print("The aws cli is required for setup.")
      print()
      print(f"if you need to install the cli you can goto here: {self.links['cli_doc_link']}\nIt is also highly recommended to install the ssm plugin here: {self.links['ssm_doc_link']}")
      print()
      print()
      print("-----------------------------")
      
      self.update_is_cli_installed(is_cli_installed= self._is_aws_installed())
      self.update_sdk_auth()
      print("cli state updated")
      print("-----------------------------")
      
      if self.is_cli_installed() != True:
        print("Please install the aws cli, if its not already installed.")
        return False

    return True

  def _is_aws_installed(self):
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "installed": {
            "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
            "messages":{
              "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}",
            },
            "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
            "desc": f"Have you already installed the aws cli?\nValid Options: {self.get_common().helper_type().bool().is_true_values()}",
            "default": self.is_cli_installed(),
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": True
        },
      }
    )

    
    return self.__get_installed(installed= response) == True

  def __get_installed(self, installed, *args, **kwargs):
    if installed is None:
      return ""
    
    if installed.get("installed") is not None:
      return self.__get_installed(installed= installed.get("installed"))
    
    return installed.get("formated") if installed.get("formated") is not None else False


from threemystic_cloud_client.cloud_providers.aws.base_class.base import cloud_client_provider_aws_base as base


class cloud_client_aws(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_aws", *args, **kwargs)
  
  # There is not post init when in Config Mode
  def _post_init(self, *args, **kwargs):
    pass
  
  def action_test(self, *args, **kwargs):
    if not self.is_provider_config_completed():
      print("Provider must be configured first")
      self._setup_another_config()
      return
    from threemystic_cloud_client.cloud_providers.aws.action_test.step_1 import cloud_client_aws_test_step_1 as test
    next_step = test(common= self.get_common(), logger= self.get_common().get_logger(), *args, **kwargs)
    
    next_step.step()

  def action_config(self, *args, **kwargs):     
    from threemystic_cloud_client.cloud_providers.aws.config.step_1 import cloud_client_aws_config_step_1 as step
    next_step = step(common= self.get_common(), logger= self.get_common().get_logger())
    
    next_step.step()

  def action_token(self, *args, **kwargs):     
    from threemystic_cloud_client.cloud_providers.aws.action_token.step_1 import cloud_client_aws_token_step_1 as step
    next_step = step(common= self.get_common(), logger= self.get_common().get_logger())
    
    next_step.step()

  def action_generate(self, *args, **kwargs):     
    from threemystic_cloud_client.cloud_providers.aws.action_generate.step_1 import cloud_client_aws_generate_step_1 as step
    next_step = step(common= self.get_common(), logger= self.get_common().get_logger())
    
    next_step.step()


  
    
    
  



from threemystic_cloud_client.cloud_providers.azure.config.base_class.base import cloud_client_azure_config_base as base

class cloud_client_azure_config_step_1(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_azure_config_step_1", *args, **kwargs)
    

  def step(self, *args, **kwargs):
    
    if not super().step(force_cli_installed_prompt= True):
      return
    
    if self.is_cli_installed() == True:
      print("Currently this app integrates directly with the azure cli. So as long as the CLI is installed the base is done.")
      print()
      self.update_provider_config_completed(status= True)
    else:
      print("The Azure CLI is required please rerun the config once the CLI has been installed.")
      print()

    self._setup_another_config()
    


import os
from abc import abstractmethod
from threemystic_cloud_client.cloud_providers.aws.base_class.base import cloud_client_provider_aws_base as base


class cloud_client_aws_client_base(base):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

    self._set_profile(*args, **kwargs)
  
  @abstractmethod
  def _session_expired(self, refresh = False, *args, **kwargs):
    pass

  @abstractmethod
  def _load_base_configs(self):
    pass
  
  @abstractmethod
  def get_main_account_id(self, *args, **kwargs):
    pass

  @abstractmethod
  def get_organization_account_id(self, *args, **kwargs):
    pass
  
  @abstractmethod
  def _assume_role(self, *args, **kwargs):    
    pass
  
  @abstractmethod
  def get_default_rolename(self, *args, **kwargs):
    pass
  
  @abstractmethod
  def get_default_region(self, *args, **kwargs):
    pass
  
  @abstractmethod
  def get_default_account(self, *args, **kwargs):
    pass
  
  @abstractmethod
  def authenticate_session(self, force_quiet = False, *args, **kwargs):
    pass
  
  def _post_init(self, *args, **kwargs):
    self._load_base_configs()
  
  


from threemystic_cloud_client.cloud_providers.aws.action_test.base_class.base import cloud_client_aws_test_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers


class cloud_client_aws_test_step_2(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_aws_test2", *args, **kwargs)
    

  def step(self, profile_name, *args, **kwargs):
    if not super().step(*args, **kwargs):
      return
    
    from threemystic_cloud_client.cloud_client import cloud_client
    aws_client = cloud_client(logger= self.get_common().get_logger(), common=self.get_common()).client(
      provider= "aws",
      profile_name= profile_name
    )

    
    print(f"Connected to Account ID: {aws_client.get_main_account_id()}")
    print(f"Org Account ID: {aws_client.get_organization_account_id()}")
    print(f"You have the following accounts:")
    for account in aws_client.get_accounts():
      print(f"{aws_client.get_account_id(account= account)}:{aws_client.get_account_name(account= account)}")
    

    
    
    
    
  



from threemystic_cloud_client.cloud_providers.azure.base_class.base import cloud_client_provider_azure_base as base

class cloud_client_azure_client_auto(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_azure_auto", *args, **kwargs)

  def _login(self, *args, **kwargs):
    pass
  
  def get_client(self, *args, **kwargs):

    from threemystic_cloud_client.cloud_providers.azure.client.cli import cloud_client_azure_client_cli as client
    return client()
  


from threemystic_cloud_client.cloud_providers.aws.action_test.base_class.base import cloud_client_aws_test_base as base
from threemystic_common.base_class.base_script_options import base_process_options
import textwrap, argparse


class cloud_client_aws_token_step_1(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_aws_token", *args, **kwargs)

    self._process_cli_args()
    
  def _process_cli_args(self, *args, **kwargs):
    process_options = base_process_options(common= self.get_common())
    token_parser_args = { 
      "--help,-h": {
        "default": False,
        "dest": "client_help",
        "help": "Display Help",
        "action": 'store_true'
      }
    }
    self._arg_parser = process_options.get_parser(
      parser_init_kwargs = {
        "prog": "3mystic_cloud_client --token -p aws",
        "formatter_class": argparse.RawDescriptionHelpFormatter,
        "description": textwrap.dedent('''\
        Requires additional settings.
          --account is required"
        '''),
        "add_help": False,
        "epilog": ""
      },
      parser_args = self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        token_parser_args,
        {
          "--account": {
            "default": None, 
            "type": str,
            "dest": "token_account",
            "help": "The AWS Account ID to generate access token information for",
            "action": 'store'
          },
          "--profile": {
            "default": None, 
            "type": str,
            "dest": "token_profile",
            "help": "The 3Mystic AWS Profile to use. If not provided the default will be used",
            "action": 'store'
          },
          "--format": {
            "default": "cli", 
            "type": str,
            "choices": ["cli", "raw", "export"],
            "dest": "token_format",
            "help": "The format the token will be returned in the options are export, cli, raw. The default is cli",
            "action": 'store'
          }
        },
      ])
    )


    processed_info = process_options.process_opts(
      parser = self._arg_parser
    )
    
    self._processed_arg_info = processed_info.get("processed_data")
  
    
  def step(self, *args, **kwargs):
    if not super().step( *args, **kwargs):
      return
    
    if (self.get_common().helper_type().bool().is_true(check_value= self._processed_arg_info.get("client_help"))):
      self._arg_parser.print_help()
      return
    
    if (self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._processed_arg_info.get("token_account"))):
      self._arg_parser.print_help()
      return

    from threemystic_cloud_client.cloud_client import cloud_client
    aws_client = cloud_client(logger= self.get_common().get_logger(), common=self.get_common()).client(
      provider= "aws",
      profile_name= self._processed_arg_info.get("token_profile")
    )

    if aws_client.session_expired():
      aws_client.authenticate_session(force_quiet= True)
    
    token_format = self.get_common().helper_type().string().set_case(string_value= self._processed_arg_info.get("token_format"), case= "lower")
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= self._processed_arg_info.get("token_format")):
      token_format = "cli"
    
    if token_format == "raw":
      print(
        self.get_common().helper_json().dumps(data= aws_client.assume_role(account= self._processed_arg_info.get("token_account")))
      )
      return
    
    if token_format == "export":
      print(
        aws_client.convert_assume_role_credentials_export(credentials= aws_client.assume_role(account= self._processed_arg_info.get("token_account")))
      )
      return
    
    if aws_client.session_expired():
      raise self.get_common().exception().exception(exception_type= "generic").exception(
        message= f"You must authenticate with the provider first. To test the connection you can run\n3mystic_cloud_client -p aws -t --profile {aws_client.get_profile().get('profile_name')}"
      )
      return 1
    
    print(
      self.get_common().helper_json().dumps(data= aws_client.convert_assume_role_credentials_cli(credentials= aws_client.assume_role(account= self._processed_arg_info.get("token_account"))))
    )
    return
    
    
  



from threemystic_cloud_client.cloud_providers.azure.base_class.base import cloud_client_provider_azure_base as base

class cloud_client_azure_token_base(base):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def _login(self, *args, **kwargs):
    pass
  
  def step(self, *args, **kwargs):
    if self.is_cli_installed() != True:
      from threemystic_cloud_client.cli import cloud_client_cli
      cloud_client_cli().process_client_action(force_action= "config")
      print("-----------------------------")
      print()
      print()
      print("Continue to token")
      print("-----------------------------")
      
    return True


from threemystic_cloud_client.cloud_providers.azure.base_class.base import cloud_client_provider_azure_base as base
from abc import abstractmethod
from io import StringIO
from knack.log import CLI_LOGGER_NAME
import logging
from azure.core.exceptions import HttpResponseError
from azure.cli.core import get_default_cli
from azure.mgmt.resource.subscriptions import SubscriptionClient as ResourceSubscriptionClient
from azure.mgmt.resourcegraph import ResourceGraphClient
from azure.mgmt.resourcegraph.models import QueryRequestOptions, QueryRequest

class cloud_client_azure_client_base(base):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  
  @abstractmethod
  def _get_tenant_credential(self, tenant = None, *args, **kwargs):
    pass
  
  def get_tenant_credential_full(self, tenant = None, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= self.get_tenant_id(tenant= tenant)):
      raise self.get_common().exception().exception(
        exception_type = "argument"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = "tenant_id",
        message = f"tenant_id cannot be null or whitespace"
      )
    
    if (self._get_credential(account_id= self.get_tenant_id(tenant= tenant)) is not None
        and len(self._get_credential(account_id= self.get_tenant_id(tenant= tenant))) > 1):
      return self._get_credential(account_id= self.get_tenant_id(tenant= tenant))
    
    credentials = self._get_tenant_credential(tenant= tenant)
    try:
      jwt_token = self.decode_jwt_token(
        token= credentials.get_token("https://graph.microsoft.com/.default").token
      )
    except Exception as err:
      jwt_token = None
      self.get_common().get_logger().exception(
        msg= "could not generate token to decode",
        extra= {
          "exception": err
        }
      )
      if self.check_request_error_login(exception= err):      
        credentials = self._login(
          tenant = tenant,
          on_login_function = lambda: self._get_tenant_credential(tenant= tenant)
        )
        try:
          jwt_token = self.decode_jwt_token(
            token= credentials.get_token("https://graph.microsoft.com/.default").token
          )
        except Exception as reattempt_err:
          jwt_token = None
          
          self.get_common().get_logger().exception(
            msg= "reattempt: could not generate token to decode",
            extra= {
              "exception": reattempt_err
            }
          )
    
    self._get_credential(account_id= self.get_tenant_id(tenant= tenant) )["credentials"] = credentials
    self._get_credential(account_id= self.get_tenant_id(tenant= tenant) )["jwt"] = jwt_token

    return self.get_tenant_credential_full(tenant= tenant, *args, **kwargs)
  
  def get_tenant_credential(self, tenant = None, *args, **kwargs):
    return self.get_tenant_credential_full(tenant= tenant, *args, **kwargs).get("credentials")
         
  def get_tenants(self, refresh = False, tenant = None, *args, **kwargs):
    if tenant is None:
      return self.__get_tenants(refresh= refresh)
    
    if self.get_common().helper_type().general().is_type(obj= tenant, type_check= str) and not self.get_common().helper_type().string().is_null_or_whitespace(string_value= tenant):
      tenant = [ acct.strip() for acct in self.get_common().helper_type().string().split(string_value= tenant) if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= acct) ]
    
    if not self.get_common().helper_type().general().is_type(obj= tenant, type_check= list):
      self.get_common().get_logger().warning(f'unknown data type for accounts {type(tenant)}, when trying to get tenants')
      return self.__get_tenants()
    
    search_tenants = [ acct for acct in tenant if not self.get_common().helper_type().string().set_case(string_value= acct, case= "lower").startswith("-") ]
    exclude_tenants = [ acct for acct in tenant if self.get_common().helper_type().string().set_case(string_value= acct, case= "lower").startswith("-") ]

    return [ acct for acct in self.__get_tenants() if f'-{self.get_tenant_id(tenant= acct)}' not in exclude_tenants and (len(search_tenants) < 1 or self.get_common().helper_type().list().find_item(data= search_tenants, filter= lambda item: item == self.get_tenant_id(tenant= acct)) is not None) ]
  
  def __get_tenants(self, refresh = False, *args, **kwargs):    
    if hasattr(self, "_tenants") and not refresh:
      return self._tenants
    
    bashCall = "az account tenant list"
    return_result = self._az_cli(bashCall, on_login_function = lambda: self.get_tenants(refresh = refresh))

    if len(return_result["error"])>0:
      if return_result.get("exit_code") != 0:
        raise Exception(return_result["error"])
      else:
        logging.warning(msg=return_result["error"])

    self._tenants = return_result.get("result")
    return self.__get_tenants()
  

  def __get_raw_accounts(self, refresh = False, *args, **kwargs):    
    if hasattr(self, "_raw_accounts") and not refresh:
      return self._raw_accounts
    
    bashCall = "az account list{}".format(" --refresh" if refresh else "")
    return_result = self._az_cli(bashCall, on_login_function = lambda: self.get_accounts(refresh = refresh))

    if len(return_result["error"])>0:
      if return_result["exit_code"] != 0:
          raise Exception(return_result["error"])
      else:
          logging.warning(msg=return_result["error"])

    self._raw_accounts = return_result["result"]
    
    return self.__get_raw_accounts()
  
  def get_group_accounts_by_tenant(self, accountList = None, refresh= False, *args, **kwargs):
    if accountList is None:
      accountList = self.get_raw_accounts(refresh= refresh *args, **kwargs)
    
    return_data = {}
    for account in accountList:
      if return_data.get(self.get_tenant_id(tenant= account, is_account= True)) is None:
        return_data[self.get_tenant_id(tenant= account, is_account= True)] = []
      
      return_data[self.get_tenant_id(tenant= account, is_account= True)].append(account)
    
    return return_data


  def get_raw_accounts(self, refresh = False, account = None, tenant = None, *args, **kwargs):
    if account is None and tenant is None:
      return self.__get_raw_accounts(refresh= refresh, *args, **kwargs)
    
    if account is None:
      account = []
    if tenant is None:
      tenant = []

    if self.get_common().helper_type().general().is_type(obj= account, type_check= str) and not self.get_common().helper_type().string().is_null_or_whitespace(string_value= account):
      account = [ self.get_account_id(account= acct) for acct in account.split(",") if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= acct) ]
    
    if self.get_common().helper_type().general().is_type(obj= tenant, type_check= str) and not self.get_common().helper_type().string().is_null_or_whitespace(string_value= tenant):
      tenant = [ self.get_tenant_id(tenant= acct) for acct in tenant.split(",") if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= acct) ]
    
    if self.get_common().helper_type().general().is_type(obj= account, type_check= dict):
      tenant = [ account ]

    if self.get_common().helper_type().general().is_type(obj= tenant, type_check= dict):
      tenant = [ tenant ]
    
    if not self.get_common().helper_type().general().is_type(obj= account, type_check= list) and not self.get_common().helper_type().general().is_type(obj= tenant, type_check= list):
      if not self.get_common().helper_type().general().is_type(obj= account, type_check= list):
        self.get_common().get_logger().warning(f'unknown data type for accounts {type(account)}, when trying to get accounts')
      if not self.get_common().helper_type().general().is_type(obj= tenant, type_check= list):
        self.get_common().get_logger().warning(f'unknown data type for tenants {type(tenant)}, when trying to get tenants')
      
      return self.__get_raw_accounts()
    
    search_accounts = [ self.get_account_id(account= acct) for acct in account if not self.get_account_id(account= acct).startswith("-") ]
    exclude_accounts = [ self.get_account_id(account= acct) for acct in account if self.get_account_id(account= acct).startswith("-") ]

    search_tenants = [ self.get_tenant_id(tenant= acct) for acct in tenant if not self.get_tenant_id(tenant= acct).startswith("-") ]
    exclude_tenants = [ self.get_tenant_id(tenant= acct) for acct in tenant if self.get_tenant_id(tenant= acct).startswith("-") ]
    
    return [ acct for acct in self.__get_raw_accounts() if 
              (len(account) > 0 and (
              f'-{self.get_account_id(account= acct)}' not in exclude_accounts and 
              (len(search_accounts) < 1 or self.get_common().helper_type().list().find_item(data= search_accounts, filter= lambda item: item == self.get_account_id(account= acct)) is not None) 
              )) and
              (len(tenant) > 0 and (
              f'-{self.get_tenant_id(tenant= acct)}' not in exclude_tenants and 
              (len(search_tenants) < 1 or self.get_common().helper_type().list().find_item(data= search_tenants, filter= lambda item: item == self.get_tenant_id(tenant= acct)) is not None) 
              ))
            ] 
  
  def __get_accounts_resourcecontainers_accounts(self, tenant, accounts,  *args, **kwargs):  
    if not self.get_common().helper_type().general().is_type(obj= accounts, type_check= list):      
      raise self.get_common().exception().exception(
        exception_type = "argument"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = "accounts",
        message = f"Argument accounts is not of type list"
      )
    
    if len(accounts) < 1:
      return None

    resource_client = ResourceGraphClient(credential= self.get_tenant_credential(tenant= tenant))
    resource_query_options = QueryRequestOptions(result_format="objectArray")
    
    resource_query = QueryRequest(subscriptions=[self.get_account_id(account= acct) for acct in accounts], query="resourcecontainers | where type == 'microsoft.resources/subscriptions'", options=resource_query_options)
    try:
      return self.sdk_request(tenant= tenant, lambda_sdk_command=lambda: resource_client.resources(resource_query))
    except HttpResponseError as err:   
               
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = "AccessDenied",
        message = f"Could not query ResourceGraphClient - {err.response.status_code}: {err.response.text}",
      )
    except Exception as err:
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = "GeneralException",
        message = f"Could not query ResourceGraphClient - {err}",
      )
    
  def __get_accounts_resourcecontainers_managementgroups(self, tenant, *args, **kwargs):  
    resource_client = ResourceGraphClient(self.get_tenant_credential(tenant= tenant))
    resource_query_options = QueryRequestOptions(result_format="objectArray")
    resource_query = QueryRequest(management_groups=[self.get_tenant_id(tenant= tenant)], query="resourcecontainers | where type == 'microsoft.resources/subscriptions'", options=resource_query_options)
    try:
      return self.sdk_request(tenant= tenant, lambda_sdk_command=lambda: resource_client.resources(resource_query))
    except HttpResponseError as err:       
      if err.status_code == 403:
        return self.__get_accounts_resourcecontainers_accounts(tenant= tenant, accounts= self.get_raw_accounts(tenant= tenant), *args, **kwargs)

      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = "AccessDenied",
        message = f"Could not query ResourceGraphClient",
        exception= err
      )
    except Exception as err:
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = "GeneralException",
        message = f"Could not query ResourceGraphClient",
        exception= err
      )
    
  def __get_accounts(self, refresh = False, *args, **kwargs):   
    
    accounts = self.get_group_accounts_by_tenant(refresh= refresh, *args, **kwargs)
    subscriptions = {}
    for tenant_id, tenant_accounts in accounts.items():
      # pulls all subscriptions to know which ones are resourcecontainers
      resource_query_results = self.__get_accounts_resourcecontainers_accounts(tenant= tenant_id, accounts= tenant_accounts)

      subscription_details = {self.get_account_id(account= account):account for account in resource_query_results.data} if resource_query_results is not None else {}
      
      resource_subscription_client = ResourceSubscriptionClient(
        (self.get_tenant_credential(tenant= tenant_id))
      )  

      subscriptions[self.get_tenant_id(tenant= tenant_id)] = [resource_subscription for resource_subscription in self.sdk_request(tenant= tenant_id, lambda_sdk_command=lambda: resource_subscription_client.subscriptions.list())]

      for subscription in subscriptions[self.get_tenant_id(tenant= tenant_id)]:
        if subscription_details.get(self.get_account_id(account= subscription)) is None:
          subscription.resource_container = False
          continue
        
        subscription.resource_container = True
    
    return self.get_common().helper_type().list().flatten(data= list(subscriptions.values()))
            
  def get_accounts(self, account= None, tenant= None, refresh= False, *args, **kwargs):      
    return self.__get_accounts(account= account, tenant= tenant, refresh= refresh, *args, **kwargs)


  def _az_cli(self, command, on_login_function = None, *args, **kwargs):

    az_cli = get_default_cli()
    stdout_buffer = StringIO()
    log_buffer = StringIO()

    az_cli_logger = logging.getLogger(CLI_LOGGER_NAME)
    az_cli_handler = logging.StreamHandler(stream=log_buffer)
    az_cli_handler.setLevel(logging.WARNING)
    
    try:
      az_cli_args = self.get_common().helper_type().string().split(string_value= command, separator="\\s")
      az_cli_args.append("--only-show-errors")

      if az_cli_args[0].lower() == "az":
        az_cli_args.pop(0)

      if az_cli_args[0].lower() != "login":
        az_cli_logger.addHandler(az_cli_handler)
      
      exit_code = az_cli.invoke(
        args= az_cli_args, 
        out_file=stdout_buffer) or 0
        
    except SystemExit as ex:
      exit_code = ex.code
    
    error_message = log_buffer.getvalue()
    if exit_code == 0:
      if az_cli_args[0].lower() == "login":
        self._clear_credential()
        if on_login_function is not None:
          # look to update this to pull a page of subscriptions and put this in some sort of loop to retry
          # thought create a validate login method.
          return on_login_function()
    
      return {
        "exit_code": exit_code,
        "result": self.get_common().helper_json().loads(data= stdout_buffer.getvalue()) if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= stdout_buffer.getvalue()) else None,
        "error": error_message
      }
    
    
    
    if self.check_request_error_login(exception= error_message):
      return self._az_cli("az login", on_login_function = on_login_function)
    
    return {
      "exit_code": exit_code,
      "result": self.get_common().helper_json().loads(data= stdout_buffer.getvalue()) if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= stdout_buffer.getvalue()) else None,
      "error": error_message
    }
  
  def get_resource_tags_as_dictionary(self, resource, *args, **kwargs):
    if resource is None:
      return None

    tags = None
    if not self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      if hasattr(resource, "extra_tags"):
        if resource.extra_tags is not None and len(resource.extra_tags) > 0:
          return resource.extra_tags
        
      if hasattr(resource, "tags"):
        tags = resource.tags
    elif self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      if resource.get("extra_tags") is not None and len(resource.get("extra_tags")) > 0:
        return resource.get("extra_tags")
      tags = resource.get("tags")

    if tags is None:
      tags = {}
      
    if "name" not in tags:
      name_options = ["display_name", "displayname", "name"]

      if not self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
        for att in dir(resource):
          if self.get_common().helper_type().string().set_case(string_value= att, case= "lower") in name_options:
            tags["name"] = getattr(resource, att).strip()
            return tags

      for att in resource.keys():
        if self.get_common().helper_type().string().set_case(string_value= att, case= "lower") in name_options:
          tags["name"] = resource.get(att).strip()
          return tags
    return tags
  


from threemystic_cloud_client.cloud_providers.base_class.base import cloud_client_provider_base as base
from azure.mgmt.costmanagement.models import TimeframeType, OperatorType, QueryColumnType, QueryDefinition, QueryTimePeriod, QueryDataset, QueryAggregation, QueryGrouping, QueryFilter, QueryComparisonExpression
from azure.core.exceptions import HttpResponseError, ClientAuthenticationError
from azure.identity._exceptions import CredentialUnavailableError
import time
import math
from abc import abstractmethod
import jwt
from polling2 import TimeoutException, poll as poll2

class cloud_client_provider_azure_base(base):
  def __init__(self, *args, **kwargs):
    self._stop_process_login()
    super().__init__(provider= "azure", *args, **kwargs)
    
    self.links = {
      "cli_doc_link": "https://learn.microsoft.com/en-us/cli/azure/install-azure-cli"
    }

  
  @abstractmethod
  def _login(self, on_login_function = None, tenant = None, *args, **kwargs):
    pass

  def login(self, *args, **kwargs):
    if self.is_login_processing():
      poll2(
        lambda: self.is_login_processing(),
        ignore_exceptions=(Exception,),
        timeout=240,
        step=0.1
      )

    self._start_process_login()
    return_data = self._login(*args, **kwargs)
    self._stop_process_login()

    return return_data
  
  def decode_jwt_token(self, token, *args, **kwargs):
    alg = jwt.get_unverified_header(jwt= token)
    return jwt.decode(
      jwt= token,
      algorithms= [alg],
      options= {"verify_signature": False}
    )

  def is_login_processing(self, *args, **kwargs):
    return (self.__is_processing_login is True)
  
  def _start_process_login(self, *args, **kwargs):
    self.__is_processing_login = True
  
  def _stop_process_login(self, *args, **kwargs):
    self.__is_processing_login = False

  def _start_process_login(self, *args, **kwargs):
    self.__is_processing_login = True

  
  def _get_credential(self, account_id, unset = False, *args, **kwargs):
    if unset is True:
      if hasattr(self, f"_credentials_{account_id}"):
        current_value = getattr(self, f"_credentials_{account_id}")
        delattr(self, f"_credentials_{account_id}")
        return current_value
      return None
      
    if hasattr(self, f"_credentials_{account_id}"):
      return getattr(self, f"_credentials_{account_id}")
    
    setattr(self, f"_credentials_{account_id}", {})
    return self._get_credential(account_id= account_id)

  def _clear_credential(self, *args, **kwargs):
    if hasattr(self, "_credentials"):
      delattr(self, "_credentials")
  
  def check_request_too_many_requests(self, exception, *args, **kwargs):
    if not self.get_common().helper_type().general().is_type(obj= exception, type_check= HttpResponseError):
      return False       
    
    if (HttpResponseError(exception)).status_code == 429:
      return True
    
    if "too many requests" in self.get_common().helper_type().string().set_case(string_value= (HttpResponseError(exception)).message, case= "lower"):
      return True
    
    return False
  
  def check_request_error_login(self, exception, *args, **kwargs):

    if self.get_common().helper_type().general().is_type(obj= exception, type_check= CredentialUnavailableError):    
      if "az login" in self.get_common().helper_type().string().set_case(string_value= (CredentialUnavailableError(exception)).message, case= "lower"):
        return True
        
    if self.get_common().helper_type().general().is_type(obj= exception, type_check= CredentialUnavailableError):    
      if (HttpResponseError(exception)).status_code is None:
        if "az login" in self.get_common().helper_type().string().set_case(string_value= (HttpResponseError(exception)).message, case= "lower"):
          return True
    
    
    if self.get_common().helper_type().general().is_type(obj= exception, type_check= Exception):    
      if "az login" in self.get_common().helper_type().string().set_case(string_value= f'{exception}', case= "lower"):
        return True
      
    if self.get_common().helper_type().general().is_type(obj= exception, type_check= str):
      exception = self.get_common().helper_type().string().set_case(
        string_value= exception, 
        case= "lower"
      )
      for exceptions_str in ["please run 'az login' to setup account"]:
        if exceptions_str in exception:
          return True
      
    # as I learn better exceptions I want to make sure I am trackign
    # if self.get_common().helper_type().general().is_type(obj= exception, type_check= ClientAuthenticationError): 
       
    
    return False
  def __get_backoff_time(self, count, *args, **kwargs):
    return self.get_common().helper_type().requests().expodential_backoff_wait(attempt= count, auto_sleep= False)

  def sdk_request(self, tenant, lambda_sdk_command, *args, **kwargs):
    max_count = 15
    count = 0
    while True:
      try:
        return lambda_sdk_command()
      except Exception as err:
        if count >= max_count:
          raise self.get_common().exception().exception(
            exception_type = "generic"
          ).not_implemented(
            logger = self.get_common().get_logger(),
            name = "sdk_request",
            message = f"too many attempts: {lambda_sdk_command}",
            exception= err
          )
        
        if(self.check_request_too_many_requests(exception= err)):
          count += 1
          time.sleep(self.__get_backoff_time(count= count))
          continue
        
        if self.check_request_error_login(exception= err):
          return self.login(tenant= tenant, on_login_function= lambda: self.sdk_request(tenant= tenant, lambda_sdk_command=lambda_sdk_command, *args, **kwargs))

        raise err
     
  def get_default_querydefinition(self):
    return QueryDefinition(
      type= "ActualCost",
      timeframe= TimeframeType.THE_LAST_MONTH,
      dataset= QueryDataset(
      aggregation= {
        "totalCost": QueryAggregation(
          name= "Cost",
          function= "Sum"
        )
      },
      grouping= [
        QueryGrouping(
          type= QueryColumnType.DIMENSION,
          name= "SubscriptionId"
        )
      ]
      )
    )
  
  def serialize_resource(self, resource, *args, **kwargs):
    if resource is None:
      return None
    
    if self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      return resource

    return resource.serialize(keep_readonly= True)
    
  def deserialize_resource(self, resource, aztype):
    if resource is None:
      return None

    return aztype.deserialize(resource)
  
  def get_resource_name(self, resource, *args, **kwargs):
    if resource is None:
      return None

    if self.get_common().helper_type().general().is_type(obj= resource, type_check= str):
      return resource
    
    if self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      return self.get_resource_name(resource= resource.get("extra_name") if not self.get_common().helper_type().string().is_null_or_whitespace(string_value=resource.get("extra_name")) else resource.get("name"))

    return self.get_resource_name(resource= getattr(resource, "name"))

  def get_resource_id_short(self, resource, include_resource_group = False, *args, **kwargs):
    resource_id = self.get_resource_id(resource= resource)
    resource_id = resource_id[resource_id.rfind("/resourcegroups/"):]

    if include_resource_group:
      return resource_id

    return resource_id.append(resource_id[resource_id.rfind("/providers/"):])

  def get_resource_id(self, resource, *args, **kwargs):
    if resource is None:
        return None
    
    if self.get_common().helper_type().general().is_type(obj= resource, type_check= str):
      return resource

    if self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      if "extra_id" in resource and resource.get("extra_id") is not None:
        return self.get_common().helper_type().string().set_case(
          string_value= resource.get("extra_id"), 
          case= "lower"
        )
      
      return self.get_common().helper_type().string().set_case(
          string_value= resource.get("id"), 
          case= "lower"
        )

    return self.get_common().helper_type().string().set_case(
      string_value=resource.id, 
      case= "lower"
    )
  
  def get_resource_group_from_resource(self, resource, *args, **kwargs):
    if resource is None:
        return None

    if self.get_common().helper_type().general().is_type(obj= resource, type_check= str):
      return resource

    if self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      if "extra_resourcegroups" in resource and resource.get("extra_resourcegroups") is not None:
        return resource.get("extra_resourcegroups")[0]
      
      return resource.get("resource_group")
      
    if hasattr(resource, "resource_group"):
      return resource.resource_group 
    
    resource_id_split = self.get_common().helper_type().string().split(
      string_value= self.get_common().helper_type().string().set_case(
        string_value= self.get_resource_id(resource= resource), 
        case= "lower"),
      separator="/",
      regex_split= False
    )
    return resource_id_split[resource_id_split.index("resourcegroups") + 1]
  
  def get_resource_location(self, resource, *args, **kwargs):
    if resource is None:
      return None

    if not self.get_common().helper_type().general().is_type(obj= resource, type_check= str):
      if self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
          return self.get_resource_location(resource= resource.get("extra_region") if not self.get_common().helper_type().string().is_null_or_whitespace(string_value=resource.get("extra_region")) else resource.get("location"))
      return self.get_resource_location(resource= getattr(resource, "location"))

    return resource
  
  def get_tenant_prefix(self, *args, **kwargs):
    return "/tenants/"

  def get_account_prefix(self, *args, **kwargs):
    return "/subscriptions/"
  
  def get_account_name(self, account, *args, **kwargs):
    if account is None:
      return None
    if self.get_common().helper_type().general().is_type(obj= account, type_check= str):
      return account.strip()
    
    name_options = ["display_name", "displayname", "name"]

    if self.get_common().helper_type().general().is_type(obj= account, type_check= dict):
      for key, value in account.items():
        if self.get_common().helper_type().string().set_case(string_value= key, case= "lower") in name_options:
          return value.strip()
    
    for att in dir(account):
      if self.get_common().helper_type().string().set_case(string_value= att, case= "lower") in name_options:
        return getattr(account, att).strip()
    
    raise self.get_common().exception().exception(
      exception_type = "generic"
    ).not_implemented(
      logger = self.get_common().get_logger(),
      name = "account",
      message = f"Unknown account object: {account}."
    )

  def get_account_id(self, account, *args, **kwargs):
    if account is None:
      return None
    if self.get_common().helper_type().general().is_type(obj= account, type_check= str):
      return self.get_common().helper_type().string().set_case(string_value= account, case= "lower").strip()
    
    id_options = ["subscriptionid", "subscription_id", "id"]
    
    if self.get_common().helper_type().general().is_type(obj= account, type_check= dict):
      for key, value in account.items():
        if self.get_common().helper_type().string().set_case(string_value= key, case= "lower") in id_options:
          return self.get_common().helper_type().string().set_case(string_value= value, case= "lower").strip() if not self.get_common().helper_type().string().set_case(string_value= value, case= "lower").startswith(self.get_account_prefix()) else self.get_common().helper_type().string().split(string_value= value, separator= "/")[-1]

    for att in dir(account):
      if self.get_common().helper_type().string().set_case(string_value= att, case= "lower") in id_options:
        account_id = getattr(account, att)
        return self.get_common().helper_type().string().set_case(string_value= account_id, case= "lower").strip() if not self.get_common().helper_type().string().set_case(string_value= account_id, case= "lower").startswith(self.get_account_prefix()) else self.get_common().helper_type().string().split(string_value= account_id, separator= "/")[-1]
    
    raise self.get_common().exception().exception(
      exception_type = "generic"
    ).not_implemented(
      logger = self.get_common().get_logger(),
      name = "account",
      message = f"Unknown account object: {account}."
    )
  
  def make_account(self, **kwargs):    
    account = {}

    for key, value in kwargs.items():
      if self.get_common().helper_type().string().set_case(string_value= key, case= "lower") == "id":
        account["Id"] = self.get_common().helper_type().string().set_case(string_value= value, case= "lower").strip()
        continue
      if self.get_common().helper_type().string().set_case(string_value= key, case= "lower") == "name":
        account["Name"] = value.strip()
        continue
        
    
    return account
  
  
  def get_tenant_id(self, tenant, is_account = False, *args, **kwargs):
    if tenant is None:
      return None
    if self.get_common().helper_type().general().is_type(obj= tenant, type_check= str):
      return self.get_common().helper_type().string().set_case(string_value= tenant, case= "lower")
    
    id_options = ["tenantid", "tenant_id", "id"] if not is_account else ["tenantid", "tenant_id"]
    
    if self.get_common().helper_type().general().is_type(obj= tenant, type_check= dict):
      for key, value in tenant.items():
        if self.get_common().helper_type().string().set_case(string_value= key, case= "lower") in id_options:
          return self.get_common().helper_type().string().set_case(string_value= value, case= "lower") if not self.get_common().helper_type().string().set_case(string_value= value, case= "lower").startswith(self.get_tenant_prefix()) else self.get_common().helper_type().string().split(string_value= value, separator= "/")[-1]

    for att in dir(tenant):
      if self.get_common().helper_type().string().set_case(string_value= att, case= "lower") in id_options:
        tenant_id = getattr(tenant, att)
        return self.get_common().helper_type().string().set_case(string_value= tenant_id, case= "lower") if not self.get_common().helper_type().string().set_case(string_value= tenant_id, case= "lower").startswith(self.get_tenant_prefix()) else self.get_common().helper_type().string().split(string_value= tenant_id, separator= "/")[-1]
    
    raise self.get_common().exception().exception(
      exception_type = "generic"
    ).not_implemented(
      logger = self.get_common().get_logger(),
      name = "tenant",
      message = f"Unknown tenant object: {tenant}."
    )
  
  def make_tenant(self, **kwargs):    
    tenant = {}

    for key, value in kwargs.items():
      if self.get_common().helper_type().string().set_case(string_value= key, case= "lower") == "id":
        tenant["Id"] = self.get_common().helper_type().string().set_case(string_value= value, case= "lower")
        continue
      
      if self.get_common().helper_type().string().set_case(string_value= key, case= "lower") == "name":
        tenant["Name"] = value
        continue
        
    
    return tenant
    




from threemystic_cloud_client.cloud_providers.azure.base_class.base import cloud_client_provider_azure_base as base

class cloud_client_azure(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_azure", *args, **kwargs)
  
  def _login(self, *args, **kwargs):
    pass
  
  def action_test(self, *args, **kwargs):    
    if not self.is_provider_config_completed():
      print("Provider must be configured first")
      self._setup_another_config()
      return
    
    from threemystic_cloud_client.cloud_providers.azure.action_test.step_1 import cloud_client_azure_test_step_1 as test
    next_step = test(common= self.get_common(), logger= self.get_common().get_logger(), *args, **kwargs)
    
    next_step.step()
  
  def action_config(self, *args, **kwargs):     
    from threemystic_cloud_client.cloud_providers.azure.config.step_1 import cloud_client_azure_config_step_1 as step
    next_step = step(common= self.get_common(), logger= self.get_common().get_logger())
    
    next_step.step()
  
  def action_token(self, *args, **kwargs):     
    from threemystic_cloud_client.cloud_providers.azure.action_token.step_1 import cloud_client_azure_token_step_1 as step
    next_step = step(common= self.get_common(), logger= self.get_common().get_logger())
    
    next_step.step()


    
    
  



from threemystic_cloud_client.cloud_providers.aws.config.base_class.base import cloud_client_aws_config_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers
from threemystic_cloud_client.cloud_providers.aws.config.step_2 import cloud_client_aws_config_step_2 as nextstep


class cloud_client_aws_config_step_1(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_aws_config_step_1", *args, **kwargs)
    

  def step(self, *args, **kwargs):
    
    if not super().step(force_cli_installed_prompt= True):
      return
    
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "type": {
            "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
            "messages":{
              "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}",
            },
            "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
            "desc": f"New Configuration\nValid optiond for yes: {self.get_common().helper_type().bool().is_true_values()}",
            "default": True,
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": True
        }
      }
    )

    if response is None:
      return

    self.update_provider_config_completed(status= "step1")
    nextstep(init_object = self).step(is_new_config= response["type"].get("formated") == True)
    
  



from threemystic_common.base_class.base_provider import base
from abc import abstractmethod
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers
from threemystic_cloud_client.cli import cloud_client_cli

class cloud_client_provider_base(base):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    
    self._post_init(*args, **kwargs)

  
  def _post_init(self, *args, **kwargs):
    pass

  def _setup_another_config(self, force_config = False, *args, **kwargs):
    if not force_config:
      response = self.get_common().generate_data().generate(
        generate_data_config = {
          "repeat_config": {
              "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
              "messages":{
                "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}",
              },
              "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
              "desc": f"Do you want to setup another provider?: {self.get_common().helper_type().bool().is_true_values()}",
              "default": None,
              "handler": generate_data_handlers.get_handler(handler= "base"),
              "optional": True
          }
        }
      )

      if response is None:
        return
      
      if response.get("repeat_config") is None:
        return
      
      if response.get("repeat_config").get("formated") is not True:
        return
    
    print()
    print()
    print("-------------------------------------------------------------------------")
    print()
    print()

    cloud_client_cli().process_client_action("config")
  
  def update_provider_config_completed(self, status, *args, **kwargs):
    self.get_config()["_config_process"] = status
    self._save_config()
  
  def is_provider_config_completed(self, *args, **kwargs):
    return self.get_config().get("_config_process") is True and self.is_cli_installed()
  
  def get_main_directory_name(self, *args, **kwargs):
    return "client"

  @abstractmethod
  def get_account_name(self, account, *args, **kwargs):
    pass

  @abstractmethod
  def get_account_id(self, account, *args, **kwargs):
    pass

  @abstractmethod
  def make_account(self, account, *args, **kwargs):
    pass  
  
  @abstractmethod
  def get_resource_group_from_resource(self, account, *args, **kwargs):
    pass
  
  def get_account_prefix(self, *args, **kwargs):
    return ""
  
  def get_resource_location(self, resource, *args, **kwargs):
    if resource is None:
      return None
    
  def serialize_resource(self, resource, *args, **kwargs):
    if resource is None:
      return None
    
    if not self.get_common().helper_type().general().is_type(obj= resource, type_check= dict):
      raise self.get_common().exception().exception(
        exception_type = "argument"
      ).type_error(
        logger = self.get_common().get_logger(),
        name = "resource",
        message = f"Unknown argument type - resource {type(resource)}"
      )
      
    return resource
    
  def __load_config(self, *args, **kwargs):
    config_data = self.get_common().helper_config().load(
      path= self.config_path(),
      config_type= "yaml"
    )
    if config_data is not None:
      return config_data
    
    return {}

  def is_cli_installed(self, *args, **kwargs):
    return self.get_config().get("cli_installed") == True

  def get_aws_poll_login(self, *args, **kwargs):
    if(self.get_config().get("aws_poll_login") is None):
      return 120
    return self.get_config().get("aws_poll_login")

  def get_aws_poll_authenticate(self, *args, **kwargs):
    if(self.get_config().get("aws_poll_authenticate") is None):
      return 240
    return self.get_config().get("aws_poll_authenticate")

  def valid_auth_options(self, *args, **kwargs):
    return ["sso"]

  def config_path(self, *args, **kwargs):
    return self.get_common().get_threemystic_directory_config().joinpath(f"{self.get_main_directory_name()}/config_{self.get_provider()}")
    
  def get_aws_user_path(self, *args, **kwargs):
    return self.get_common().helper_path().expandpath_user("~/.aws")
  
  def get_aws_user_path_config(self, *args, **kwargs):
    return f"{self.get_aws_user_path()}/config"
  
  def get_default_profile_name(self, *args, **kwargs):

    default_profile = self.get_default_profile()
    if(default_profile is None):
      return None
    
    return default_profile["profile_name"]
  
  
  def has_default_profile(self, profile_name = None, *args, **kwargs):
    return self.get_default_profile() != None

  def get_default_profile(self, *args, **kwargs):
      
    for profile, profile_data in self.get_config_profiles().items():
      if(not profile_data["default_profile"]):
        continue
      
      return {
        "profile_name": self.get_common().helper_type().string().set_case(string_value= profile, case= "lower"),
        "profile_data": profile_data
      }
    
    return None

  def config_profile_name_exists(self, *args, **kwargs):
    
    return self.get_config_profile_name(*args, **kwargs) != None
  
  def _update_config(self,config_key, config_value, refresh= False,  *args, **kwargs):
     self.get_config(refresh = True)[config_key] = config_value
  
  def _save_config(self, *args, **kwargs):
     if not self.config_path().parent.exists():
       self.config_path().parent.mkdir(parents= True)
     self.config_path().write_text(
      data= self.get_common().helper_yaml().dumps(data= self.get_config())
     )
     self.get_config(refresh = True)

  def get_config(self, refresh = False, *args, **kwargs):
    if hasattr(self, "_config_data") and not refresh:
      return self._config_data
    
    self._config_data = self.__load_config()    
    self._config_data["profiles"] = {self.get_common().helper_type().string().set_case(string_value= profile_name, case= "lower"):profile_data for profile_name,profile_data in self.get_config_profiles().items()}

    return self.get_config(*args, **kwargs)
  
  def get_config_value(self, config_key, default_if_none = None, refresh = False, *args, **kwargs):
    config_value = self.get_config(refresh= refresh).get(config_key)
    if config_value is not None:
      return config_value
    
    return default_if_none

  def has_config_profiles(self, *args, **kwargs):
    
    if self.get_config().get("profiles") is None:
      return False
    
    return len(self.get_config().get("profiles")) > 0

  def get_config_profiles(self, *args, **kwargs):
    if self.get_config().get("profiles") is not None:
      return self.get_config().get("profiles")
    
    self.get_config()["profiles"] = {}
    return self.get_config_profiles()

  def get_config_profile_name(self, profile_name = None, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_name):
      return None
    
    profile_name = self.get_common().helper_type().string().set_case(string_value= profile_name, case= "lower")
    for existing_profile_name, profile_data in self.get_config_profiles().items():
      if(existing_profile_name != profile_name):
        continue
      
      return profile_data
    
    return None

  
  def action_config(self, *args, **kwargs):
    print("Provider config not configured")
  
  def action_test(self, *args, **kwargs):
    print("Provider test config not configured")
  
  def action_token(self, *args, **kwargs):
    print("Provider token config not configured")

  def action_generate(self, *args, **kwargs):     
    print("Provider generate config not configured/not needed")

    

  
  




from threemystic_cloud_client.cloud_providers.aws.base_class.base import cloud_client_provider_aws_base as base

class cloud_client_aws_client_auto(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_client_aws_auto", *args, **kwargs)

    self._set_profile(*args, **kwargs)
  
  def get_resource_group_from_resource(self, account, *args, **kwargs):
    pass
  
  def get_profile(self, *args, **kwargs):
    if(not hasattr(self, "_profile")):
      raise self.get_common().exception().exception(
          exception_type = "generic"
        ).type_error(
          logger = self.get_common().get_logger(),
          name = "Cloud Client Profile",
          message = f"Profile was not set"
        )
    
    if(self._profile is None):
      raise self.get_common().exception().exception(
          exception_type = "generic"
        ).type_error(
          logger = self.get_common().get_logger(),
          name = "Cloud Client Profile",
          message = f"Profile is None"
        )
    
    return self._profile
  
  def _set_profile(self, profile_name = None, *args, **kwargs):
    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= profile_name):
      profile_data = self.get_config_profile_name(profile_name= profile_name)
      if profile_data is None:
        raise self.get_common().exception().exception(
          exception_type = "argument"
        ).type_error(
          logger = self.get_common().get_logger(),
          name = "profile_name",
          message = f"profile_name is not found"
        )
      
      self._profile = {
        "profile_name": profile_name,
        "profile_data": profile_data
      }
      return
    
    self._profile = self.get_default_profile()

  def get_client(self, *args, **kwargs):
    if self.get_common().helper_type().string().set_case(string_value= self.get_profile()["profile_data"]["auth_method"], case= "lower") == "sso":
      from threemystic_cloud_client.cloud_providers.aws.client.sso import cloud_client_aws_client_sso as client
      return client(profile_data= self.get_profile())
  


import sys
from abc import abstractmethod
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers


class cloud_client_action_base():
  def __init__(self, cloud_client = None, *args, **kwargs):
    self._cloud_client = cloud_client 
    if self._cloud_client is None:
      from threemystic_cloud_client.cloud_client import cloud_client
      self._cloud_client = cloud_client()
    
    
    

  def main(self, provider = None, *args, **kwargs):
    if self._cloud_client.get_common().helper_type().string().is_null_or_whitespace(string_value= provider):
      response = self._cloud_client.get_common().generate_data().generate(
        generate_data_config = {
          "provider": {
              "validation": lambda item: self._cloud_client.get_common().helper_type().string().trim(self._cloud_client.get_common().helper_type().string().set_case(string_value= item, case= "lower")) in self._cloud_client.get_supported_providers(),
              "messages":{
                "validation": f"Valid Provider Options: {self._cloud_client.get_supported_providers()}",
              },
              "conversion": lambda item: self._cloud_client.get_common().helper_type().string().trim(string_value= self._cloud_client.get_common().helper_type().string().set_case(string_value= item, case= "lower")) if item is not None else None,
              "desc": f"Which provider?\nTo not see this prompt please use the --provider/-p flag.\nvalid options are {self._cloud_client.get_supported_providers()}",
              "default": None,
              "handler": generate_data_handlers.get_handler(handler= "base"),
              "optional": True
          }
        }
      )
      
      provider = self._cloud_client.get_common().helper_type().string().set_case(string_value= self.__get_provider(response), case= "lower")
    if provider not in self._cloud_client.get_supported_providers():
      return
    
    if provider == "azure":
      self._process_provider_azure()

    
    if provider == "aws":
      self._process_provider_aws()
  
  def provider_config_status(self, provider = None, *args, **kwargs):
    provider = self._cloud_client.get_common().helper_type().string().set_case(string_value= provider, case= "lower")

    if provider not in self._cloud_client.get_supported_providers():
      return
    
    if provider == "azure":
      from threemystic_cloud_client.cloud_providers.azure import cloud_client_azure as client
      return client(common= self._cloud_client).is_provider_config_completed()

    
    if provider == "aws":
      from threemystic_cloud_client.cloud_providers.aws  import cloud_client_aws as client
      return client(common= self._cloud_client).is_provider_config_completed()
    
    return None

  @abstractmethod
  def _process_provider_aws(self, *args, **kwargs):
    pass

  @abstractmethod
  def _process_provider_azure(self, *args, **kwargs):
    pass

  def __get_provider(self, provider, *args, **kwargs):
    if provider is None:
      return ""

    if provider.get("provider") is not None:
      return self.__get_provider(provider= provider.get("provider"))
    
    return provider.get("formated") if provider.get("formated") is not None else ""
      

  
